# fsresize.py
# Filesystem resizing classes.
#
# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Anne Mulhern <amulhern@redhat.com>

import abc
import os
import tempfile

from contextlib import contextmanager
from six import add_metaclass

from ..errors import FSError
from ..size import B, KiB, MiB, GiB, KB, MB, GB
from ..import util

from . import availability
from . import task
from . import fstask
from . import dfresize

import logging
log = logging.getLogger("blivet")


@add_metaclass(abc.ABCMeta)
class FSResizeTask(fstask.FSTask):
    """ The abstract properties that any resize task must have. """

    size_fmt = abc.abstractproperty(doc="Size format string.")


@add_metaclass(abc.ABCMeta)
class FSResize(task.BasicApplication, FSResizeTask):

    """ An abstract class for resizing a filesystem. """

    description = "resize filesystem"

    args = abc.abstractproperty(doc="Resize arguments.")

    # IMPLEMENTATION methods

    @abc.abstractmethod
    def size_spec(self):
        """ Returns a string specification for the target size of the command.
            :returns: size specification
            :rtype: str
        """
        raise NotImplementedError()

    def _resize_command(self):
        return [str(self.ext)] + self.args

    def do_task(self):  # pylint: disable=arguments-differ
        """ Resize the device.

            :raises FSError: on failure
        """
        error_msgs = self.availability_errors
        if error_msgs:
            raise FSError("\n".join(error_msgs))

        try:
            ret = util.run_program(self._resize_command())
        except OSError as e:
            raise FSError(e)

        if ret:
            raise FSError("resize failed: %s" % ret)


class Ext2FSResize(FSResize):
    ext = availability.RESIZE2FS_APP
    unit = MiB

    # No bytes specification is described in the man pages. A number without
    # any suffix is interpreted as indicating the number of filesystem blocks.
    # A suffix of "s" specifies a 512 byte sector. It is omitted here because
    # the lookup is only by standard binary units.
    size_fmt = {KiB: "%dK", MiB: "%dM", GiB: "%dG"}[unit]

    def size_spec(self):
        return self.size_fmt % self.fs.target_size.convert_to(self.unit)

    @property
    def args(self):
        return ["-p", self.fs.device, self.size_spec()]


class NTFSResize(FSResize):
    ext = availability.NTFSRESIZE_APP
    unit = B
    size_fmt = {B: "%d", KB: "%dK", MB: "%dM", GB: "%dG"}[unit]

    def size_spec(self):
        return self.size_fmt % self.fs.target_size.convert_to(self.unit)

    @property
    def args(self):
        return [
            "-ff",  # need at least two 'f's to fully suppress interaction
            "-s", self.size_spec(),
            self.fs.device
        ]


class XFSResize(FSResize):
    ext = availability.XFSRESIZE_APP
    unit = B
    size_fmt = None

    @contextmanager
    def _do_temp_mount(self):
        if self.fs.status:
            yield
        else:
            dev_name = os.path.basename(self.fs.device)
            tmpdir = tempfile.mkdtemp(prefix="xfs-tempmount-%s" % dev_name)
            log.debug("mounting XFS on '%s' to '%s' for resize", self.fs.device, tmpdir)
            try:
                self.fs.mount(mountpoint=tmpdir)
            except FSError as e:
                raise FSError("Failed to mount XFS filesystem for resize: %s" % str(e))

            try:
                yield
            finally:
                util.umount(mountpoint=tmpdir)
                os.rmdir(tmpdir)

    def _get_block_size(self):
        if self.fs._current_info:
            # this should be set by update_size_info()
            for line in self.fs._current_info.split("\n"):
                if line.startswith("blocksize ="):
                    return int(line.split("=")[-1])

        raise FSError("Failed to get XFS filesystem block size for resize")

    def size_spec(self):
        # size for xfs_growfs is in blocks
        return str(self.fs.target_size.convert_to(self.unit) / self._get_block_size())

    @property
    def args(self):
        return [self.fs.system_mountpoint, "-D", self.size_spec()]

    def do_task(self):  # pylint: disable=arguments-differ
        """ Resizes the XFS format. """

        with self._do_temp_mount():
            super(XFSResize, self).do_task()


class TmpFSResize(FSResize):

    ext = availability.MOUNT_APP
    unit = MiB
    size_fmt = {KiB: "%dk", MiB: "%dm", GiB: "%dg"}[unit]

    def size_spec(self):
        return "size=%s" % (self.size_fmt % self.fs.target_size.convert_to(self.unit))

    @property
    def args(self):
        # This is too closely mixed in w/ TmpFS object, due to the
        # fact that resizing is done by mounting and that the options are
        # therefore mount options. The situation is hard to avoid, though.
        opts = self.fs.mountopts or ",".join(self.fs._mount.options)
        options = ("remount", opts, self.size_spec())
        return ['-o', ",".join(options), self.fs._type, self.fs.system_mountpoint]


class UnimplementedFSResize(dfresize.UnimplementedDFResize, FSResizeTask):

    @property
    def size_fmt(self):
        raise NotImplementedError()



# Check a .po file for basic usability
#
# This will test that the file is well-formed and that the Plural-Forms value
# is parseable
#
# Copyright (C) 2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): David Shea <dshea@redhat.com>

import gettext
import tempfile
import polib
import subprocess

def test_usability(pofile):
    # Use polib to write a mofile
    with tempfile.NamedTemporaryFile(mode="w+b") as mofile:
        pofile = polib.pofile(pofile)
        pofile.save_as_mofile(mofile.name)

        # Try to open it
        _t = gettext.GNUTranslations(fp=mofile)

def test_msgfmt(pofile):
    # Check that the .po file can actually be compiled
    with tempfile.NamedTemporaryFile(mode="w+b", suffix=".mo") as mofile:
        try:
            # Ignore the output on success
            subprocess.check_output(["msgfmt", "-c", "--verbose", "-o", mofile.name, pofile],
                                    stderr=subprocess.STDOUT, universal_newlines=True)
        except subprocess.CalledProcessError as e:
            raise AssertionError("Unable to compile %s: %s" % (pofile, e.output)) from e



# disklabel.py
# Device format classes for anaconda's storage configuration module.
#
# Copyright (C) 2009  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Dave Lehman <dlehman@redhat.com>
#

import gi
import os

gi.require_version("BlockDev", "3.0")
from gi.repository import BlockDev as blockdev

from ..storage_log import log_exception_info, log_method_call
import parted
import _ped
from ..errors import DiskLabelCommitError, InvalidDiskLabelError, AlignmentError, PartitioningError
from .. import arch
from ..events.manager import event_manager
from .. import udev
from .. import util
from ..flags import flags
from ..i18n import _, N_
from . import DeviceFormat, register_device_format
from ..size import Size

import logging
log = logging.getLogger("blivet")


class DiskLabel(DeviceFormat):

    """ Disklabel """
    _type = "disklabel"
    _name = N_("partition table")
    _formattable = True                # can be formatted
    _default_label_type = None

    def __init__(self, **kwargs):
        """
            :keyword device: full path to the block device node
            :type device: str
            :keyword str uuid: disklabel UUID
            :keyword label_type: type of disklabel to create
            :type label_type: str
            :keyword exists: whether the formatting exists
            :type exists: bool
        """
        log_method_call(self, **kwargs)
        DeviceFormat.__init__(self, **kwargs)

        self._label_type = ""
        if not self.exists:
            self._label_type = kwargs.get("label_type") or ""

        self._size = Size(0)

        self._parted_device = None
        self._parted_disk = None
        self._orig_parted_disk = None
        self._supported = True

        self._disk_label_alignment = None
        self._minimal_alignment = None
        self._optimal_alignment = None

        if self.parted_device:
            # set up the parted objects and raise exception on failure
            try:
                self.update_orig_parted_disk()
            except Exception as e:  # pylint: disable=broad-except
                self._supported = False
                self._label_type = kwargs.get("label_type") or ""
                log.warning("error setting up disklabel object on %s: %s", self.device, str(e))

    def __deepcopy__(self, memo):
        """ Create a deep copy of a Disklabel instance.

            We can't do copy.deepcopy on parted objects, which is okay.
        """
        return util.variable_copy(self, memo,
                                  shallow=('_parted_device', '_optimal_alignment', '_minimal_alignment',
                                           '_disk_label_alignment'),
                                  duplicate=('_parted_disk', '_orig_parted_disk'))

    def __repr__(self):
        s = DeviceFormat.__repr__(self)
        if flags.testing:
            return s
        s += ("  type = %(type)s  partition count = %(count)s"
              "  sector_size = %(sector_size)s\n"
              "  align_offset = %(offset)s  align_grain = %(grain)s\n"
              "  parted_disk = %(disk)s\n"
              "  orig_parted_disk = %(orig_disk)r\n"
              "  parted_device = %(dev)s\n" %
              {"type": self.label_type, "count": len(self.partitions),
               "sector_size": self.sector_size,
               "offset": self.get_alignment().offset,
               "grain": self.get_alignment().grainSize,
               "disk": self.parted_disk, "orig_disk": self._orig_parted_disk,
               "dev": self.parted_device})
        return s

    @property
    def desc(self):
        return "%s %s" % (self.label_type, self.type)

    @property
    def dict(self):
        d = super(DiskLabel, self).dict
        if flags.testing:
            return d

        d.update({"label_type": self.label_type,
                  "partition_count": len(self.partitions),
                  "sector_size": self.sector_size,
                  "offset": self.get_alignment().offset,
                  "grain_size": self.get_alignment().grainSize})
        return d

    @property
    def supported(self):
        return self._supported

    def update_parted_disk(self):
        """ re-read the disklabel from the device """
        self._parted_disk = None
        mask = event_manager.add_mask(device=os.path.basename(self.device), partitions=True)
        self.update_orig_parted_disk()
        udev.settle()
        event_manager.remove_mask(mask)

    def update_orig_parted_disk(self):
        self._orig_parted_disk = self.parted_disk.duplicate()

    def reset_parted_disk(self):
        """ Set this instance's parted_disk to reflect the disk's contents. """
        log_method_call(self, device=self.device)
        self._parted_disk = self._orig_parted_disk

    def fresh_parted_disk(self):
        """ Return a new, empty parted.Disk instance for this device. """
        log_method_call(self, device=self.device, label_type=self.label_type)
        return parted.freshDisk(device=self.parted_device, ty=self.label_type)

    @property
    def parted_disk(self):
        if not self.parted_device:
            return None

        if not self._parted_disk and self.supported:
            if self.exists:
                try:
                    self._parted_disk = parted.Disk(device=self.parted_device)
                except (_ped.DiskLabelException, _ped.IOException, NotImplementedError):
                    self._supported = False
                    return None

                if self._parted_disk.type == "loop":
                    # When the device has no partition table but it has a FS,
                    # it will be created with label type loop.  Treat the
                    # same as if the device had no label (cause it really
                    # doesn't).
                    raise InvalidDiskLabelError()
            else:
                self._parted_disk = self.fresh_parted_disk()

            # turn off cylinder alignment
            if self._parted_disk.isFlagAvailable(parted.DISK_CYLINDER_ALIGNMENT):
                self._parted_disk.unsetFlag(parted.DISK_CYLINDER_ALIGNMENT)

            # Set the boot flag on the GPT PMBR, this helps some BIOS systems boot
            if self._parted_disk.isFlagAvailable(parted.DISK_GPT_PMBR_BOOT):
                # MAC can boot as EFI or as BIOS, neither should have PMBR boot set
                if arch.is_efi() or arch.is_mactel():
                    self._parted_disk.unsetFlag(parted.DISK_GPT_PMBR_BOOT)
                    log.debug("Clear pmbr_boot on %s", self._parted_disk)
                else:
                    self._parted_disk.setFlag(parted.DISK_GPT_PMBR_BOOT)
                    log.debug("Set pmbr_boot on %s", self._parted_disk)
            else:
                log.debug("Did not change pmbr_boot on %s", self._parted_disk)

            udev.settle(quiet=True)
        return self._parted_disk

    @property
    def parted_device(self):
        if not self._parted_device and self.device:
            if os.path.exists(self.device):
                # We aren't guaranteed to be able to get a device.  In
                # particular, built-in USB flash readers show up as devices but
                # do not always have any media present, so parted won't be able
                # to find a device.
                try:
                    self._parted_device = parted.Device(path=self.device)
                except (_ped.IOException, _ped.DeviceException) as e:
                    log.error("DiskLabel.parted_device: Parted exception: %s", e)
            else:
                log.info("DiskLabel.parted_device: %s does not exist", self.device)

        if not self._parted_device:
            log.info("DiskLabel.parted_device returning None")
        return self._parted_device

    @classmethod
    def get_platform_label_types(cls):
        label_types = ["msdos", "gpt"]
        if arch.is_pmac():
            label_types = ["mac"]
        # always prefer gpt on aarch64, x86_64, and EFI plats except 32-bit ARM
        elif arch.is_aarch64() or arch.is_x86(bits=64) or (arch.is_efi() and not arch.is_arm()):
            label_types = ["gpt", "msdos"]
        elif arch.is_s390():
            label_types += ["dasd"]

        return label_types

    @classmethod
    def set_default_label_type(cls, labeltype):
        cls._default_label_type = labeltype
        log.debug("default disklabel has been set to %s", labeltype)

    def _label_type_size_check(self, label_type):
        if self.parted_device is None:
            return False

        label = parted.freshDisk(device=self.parted_device, ty=label_type)
        return self.parted_device.length < label.maxPartitionStartSector

    def _get_best_label_type(self):
        label_type = self._default_label_type
        label_types = self.get_platform_label_types()[:]
        if label_type in label_types:
            label_types.remove(label_type)
        if label_type:
            label_types.insert(0, label_type)

        if arch.is_s390():
            if blockdev.s390.dasd_is_fba(self.device):
                # the device is FBA DASD
                return "msdos"
            elif self.parted_device.type == parted.DEVICE_DASD:
                # the device is DASD
                return "dasd"
            elif util.detect_virt():
                # check for dasds exported into qemu as normal virtio/scsi disks
                try:
                    _parted_disk = parted.Disk(device=self.parted_device)
                except (_ped.DiskLabelException, _ped.IOException, NotImplementedError):
                    pass
                else:
                    if _parted_disk.type == "dasd":
                        return "dasd"

        for lt in label_types:
            if self._label_type_size_check(lt):
                log.debug("selecting %s disklabel for %s based on size",
                          label_type, os.path.basename(self.device))
                label_type = lt
                break

        return label_type

    @property
    def label_type(self):
        """ The disklabel type (eg: 'gpt', 'msdos') """
        if not self.supported:
            return self._label_type

        # For new disklabels, user-specified type overrides built-in logic.
        # XXX This determines the type we pass to parted.Disk
        if not self.exists and not self._parted_disk:
            if self._label_type:
                lt = self._label_type
            else:
                try:
                    lt = self._get_best_label_type()
                except Exception:  # pylint: disable=broad-except
                    log_exception_info()
                    lt = self._label_type
            return lt

        try:
            lt = self.parted_disk.type
        except Exception:  # pylint: disable=broad-except
            log_exception_info()
            lt = self._label_type
        return lt

    @property
    def sector_size(self):
        try:
            return Size(self.parted_device.sectorSize)
        except AttributeError:
            log_exception_info()
            return None

    @property
    def name(self):
        if not self.label_type:
            return "%(name)s" % {"name": _(self._name)}

        if self.supported:
            return "%(name)s (%(type)s)" % {"name": _(self._name), "type": self.label_type.upper()}
        else:
            # Translators: Name for an unsupported disklabel; e.g. "Unsupported partition table"
            return _("Unsupported %(name)s") % {"name": _(self._name)}

    @property
    def size(self):
        size = self._size
        if not size:
            try:
                size = Size(self.parted_device.getLength(unit="B"))
            except Exception:  # pylint: disable=broad-except
                log_exception_info()
                size = Size(0)

        return size

    @property
    def status(self):
        """ Device status. """
        return False

    @property
    def supports_names(self):
        if not self.supported or not self.parted_disk:
            return False

        return self.parted_disk.supportsFeature(parted.DISK_TYPE_PARTITION_NAME)

    def _create(self, **kwargs):
        """ Create the device. """
        log_method_call(self, device=self.device,
                        type=self.type, status=self.status)
        # We're relying on someone having called reset_parted_disk -- we
        # could ensure a fresh disklabel by setting self._parted_disk to
        # None right before calling self.commit(), but that might hide
        # other problems.
        self.commit()

    def commit(self):
        """ Commit the current partition table to disk and notify the OS. """
        log_method_call(self, device=self.device,
                        numparts=len(self.partitions))
        try:
            self.parted_disk.commit()
        except parted.DiskException as msg:
            raise DiskLabelCommitError(msg)
        else:
            self.update_orig_parted_disk()
            udev.settle()

    def commit_to_disk(self):
        """ Commit the current partition table to disk. """
        log_method_call(self, device=self.device,
                        numparts=len(self.partitions))
        try:
            self.parted_disk.commitToDevice()
        except parted.DiskException as msg:
            raise DiskLabelCommitError(msg)
        else:
            self.update_orig_parted_disk()

    def add_partition(self, start, end, ptype=None, part_type_uuid=None):
        """ Add a partition to the disklabel.

            :param int start: start sector
            :param int end: end sector
            :param ptype: partition type or None
            :type ptype: int (parted partition type constant) or NoneType
            :param part_type_uuid: partition type UUID or None
            :type part_type_uuid: uuid.UUID or NoneType
            :raises :class:`~.errors.PartitioningError`: if parted is too old to
                                                         support partition type UUIDs

            Partition type will default to either PARTITION_NORMAL or
            PARTITION_LOGICAL, depending on whether the start sector is within
            an extended partition.
        """
        if ptype is None:
            extended = self.extended_partition
            if extended and extended.geometry.contains(start):
                ptype = parted.PARTITION_LOGICAL
            else:
                ptype = parted.PARTITION_NORMAL

        geometry = parted.Geometry(device=self.parted_device,
                                   start=start, end=end)
        new_partition = parted.Partition(disk=self.parted_disk,
                                         type=ptype,
                                         geometry=geometry)

        if part_type_uuid is not None:
            if not hasattr(parted.Partition, 'type_uuid'):
                raise PartitioningError("parted too old to set partition type UUID")
            new_partition.type_uuid = part_type_uuid.bytes

        constraint = parted.Constraint(exactGeom=geometry)
        self.parted_disk.addPartition(partition=new_partition,
                                      constraint=constraint)

    def remove_partition(self, partition):
        """ Remove a partition from the disklabel.

            :param partition: the partition to remove
            :type partition: :class:`parted.Partition`
        """
        self.parted_disk.removePartition(partition)

    @property
    def extended_partition(self):
        try:
            extended = self.parted_disk.getExtendedPartition()
        except Exception:  # pylint: disable=broad-except
            log_exception_info()
            extended = None
        return extended

    @property
    def logical_partitions(self):
        try:
            logicals = self.parted_disk.getLogicalPartitions()
        except Exception:  # pylint: disable=broad-except
            log_exception_info()
            logicals = []
        return logicals

    @property
    def primary_partitions(self):
        try:
            primaries = self.parted_disk.getPrimaryPartitions()
        except Exception:  # pylint: disable=broad-except
            log_exception_info()
            primaries = []
        return primaries

    @property
    def first_partition(self):
        try:
            part = self.parted_disk.getFirstPartition()
        except Exception:  # pylint: disable=broad-except
            log_exception_info()
            part = None
        return part

    @property
    def partitions(self):
        return getattr(self.parted_disk, "partitions", [])

    def _get_disk_label_alignment(self):
        """ Return the disklabel's required alignment for new partitions.

            :rtype: :class:`parted.Alignment`
        """
        if not self._disk_label_alignment:
            try:
                self._disk_label_alignment = self.parted_disk.partitionAlignment
            except (_ped.CreateException, AttributeError):
                self._disk_label_alignment = parted.Alignment(offset=0,
                                                              grainSize=1)

        return self._disk_label_alignment

    def get_minimal_alignment(self):
        """ Return the device's minimal alignment for new partitions.

            :rtype: :class:`parted.Alignment`
        """
        if not self._minimal_alignment:
            disklabel_alignment = self._get_disk_label_alignment()
            try:
                minimal_alignment = self.parted_device.minimumAlignment
            except (_ped.CreateException, AttributeError):
                # handle this in the same place we'd handle an ArithmeticError
                minimal_alignment = None

            try:
                alignment = minimal_alignment.intersect(disklabel_alignment)
            except (ArithmeticError, AttributeError):
                alignment = disklabel_alignment

            self._minimal_alignment = alignment

        return self._minimal_alignment

    def get_optimal_alignment(self):
        """ Return the device's optimal alignment for new partitions.

            :rtype: :class:`parted.Alignment`

            .. note::

                If there is no device-supplied optimal alignment this method
                returns the minimal device alignment.
        """
        if not self._optimal_alignment:
            disklabel_alignment = self._get_disk_label_alignment()
            try:
                optimal_alignment = self.parted_device.optimumAlignment
            except (_ped.CreateException, AttributeError):
                # if there is no optimal alignment, use the minimal alignment,
                # which has already been intersected with the disklabel
                # alignment
                alignment = self.get_minimal_alignment()
            else:
                try:
                    alignment = optimal_alignment.intersect(disklabel_alignment)
                except ArithmeticError:
                    alignment = disklabel_alignment

            self._optimal_alignment = alignment

        return self._optimal_alignment

    def get_alignment(self, size=None):
        """ Return an appropriate alignment for a new partition.

            :keyword size: proposed partition size (optional)
            :type size: :class:`~.size.Size`
            :returns: the appropriate alignment to use
            :rtype: :class:`parted.Alignment`
            :raises :class:`~.errors.AlignmentError`: if the partition is too
                                                         small to be aligned
        """
        # default to the optimal alignment
        alignment = self.get_optimal_alignment()
        if size is None:
            return alignment

        # use the minimal alignment if the requested size is smaller than the
        # optimal io size
        minimal_alignment = self.get_minimal_alignment()
        optimal_grain_size = Size(alignment.grainSize * self.sector_size)
        minimal_grain_size = Size(minimal_alignment.grainSize * self.sector_size)
        if size < minimal_grain_size:
            raise AlignmentError("requested size cannot be aligned")
        elif size < optimal_grain_size:
            alignment = minimal_alignment

        return alignment

    def get_end_alignment(self, size=None, alignment=None):
        """ Return an appropriate end-alignment for a new partition.

            :keyword size: proposed partition size (optional)
            :type size: :class:`~.size.Size`
            :keyword alignment: the start alignment (optional)
            :type alignment: :class:`parted.Alignment`
            :returns: the appropriate alignment to use
            :rtype: :class:`parted.Alignment`
            :raises :class:`~.errors.AlignmentError`: if the partition is too
                                                         small to be aligned
        """
        if alignment is None:
            alignment = self.get_alignment(size=size)

        return parted.Alignment(offset=alignment.offset - 1,
                                grainSize=alignment.grainSize)

    @property
    def alignment(self):
        return self.get_alignment()

    @property
    def end_alignment(self):
        return self.get_end_alignment()

    @property
    def free(self):
        if self.parted_disk is not None:
            free_areas = self.parted_disk.getFreeSpacePartitions()
        else:
            free_areas = []

        return sum((Size(f.getLength(unit="B")) for f in free_areas), Size(0))

    @property
    def magic_partition_number(self):
        """ Number of disklabel-type-specific special partition. """
        if self.label_type == "mac":
            return 1
        elif self.label_type == "sun":
            return 3
        else:
            return 0


register_device_format(DiskLabel)



# populator/helpers/stratis.py
# Stratis backend code for populating a DeviceTree.
#
# Copyright (C) 2020  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): Vojtech Trefny <vtrefny@redhat.com>
#

import copy

from ...callbacks import callbacks
from ... import udev
from ...formats import get_format
from ...devices.stratis import StratisPoolDevice, StratisFilesystemDevice
from ...devicelibs.stratis import STRATIS_FS_SIZE
from ...storage_log import log_method_call
from .formatpopulator import FormatPopulator

from ...static_data import stratis_info

import logging
log = logging.getLogger("blivet")


class StratisFormatPopulator(FormatPopulator):
    priority = 100
    _type_specifier = "stratis"

    @classmethod
    def match(cls, data, device):  # pylint: disable=arguments-differ,unused-argument
        if super(StratisFormatPopulator, cls).match(data, device):
            return True

        # locked stratis pools are managed here
        for pool in stratis_info.locked_pools:
            if device.path in pool.devices:
                return True

        # unlocked encrypted pools are also managed here
        if udev.device_get_format(data) == "crypto_LUKS":
            holders = udev.device_get_holders(data)
            if holders:
                fs = udev.device_get_format(holders[0])
                if fs == cls._type_specifier:
                    return True

        return False

    def _get_blockdev_uuid(self):
        if udev.device_get_format(self.data) == "crypto_LUKS":
            holders = udev.device_get_holders(self.data)
            if holders:
                return udev.device_get_uuid(holders[0])

        return udev.device_get_uuid(self.data)

    def _get_kwargs(self):
        kwargs = super(StratisFormatPopulator, self)._get_kwargs()

        name = udev.device_get_name(self.data)
        uuid = self._get_blockdev_uuid()
        kwargs["uuid"] = uuid

        # stratis block device hosting an encrypted pool
        kwargs["locked_pool"] = False
        for pool in stratis_info.locked_pools:
            if self.device.path in pool.devices:
                kwargs["locked_pool"] = True
                kwargs["pool_uuid"] = pool.uuid
                kwargs["locked_pool_key_desc"] = pool.key_desc
                return kwargs

        bd_info = stratis_info.blockdevs.get(uuid)
        if bd_info:
            if bd_info.pool_name:
                kwargs["pool_name"] = bd_info.pool_name
            else:
                log.warning("Stratis block device %s has no pool_name", name)
            if bd_info.pool_uuid:
                kwargs["pool_uuid"] = bd_info.pool_uuid
            else:
                log.warning("Stratis block device %s has no pool_uuid", name)

        return kwargs

    def _add_pool_device(self):
        bd_info = stratis_info.blockdevs.get(self.device.format.uuid)
        if not bd_info:
            # no info about the stratis block device -> we're done
            return

        if not bd_info.pool_name:
            log.info("stratis block device %s has no pool", self.device.name)
            return

        pool_info = stratis_info.pools.get(bd_info.pool_uuid)
        if pool_info is None:
            log.warning("Failed to get information about Stratis pool %s (%s)",
                        bd_info.pool_name, bd_info.pool_uuid)
            return

        pool_device = self._devicetree.get_device_by_uuid(bd_info.pool_uuid)
        if pool_device and self.device not in pool_device.parents:
            pool_device.parents.append(self.device)
            callbacks.parent_added(device=pool_device, parent=self.device)
        elif pool_device is None:
            # TODO: stratis duplicate pool name

            pool_device = StratisPoolDevice(pool_info.name,
                                            parents=[self.device],
                                            uuid=pool_info.uuid,
                                            size=pool_info.physical_size,
                                            exists=True,
                                            encrypted=pool_info.encrypted)
            self._devicetree._add_device(pool_device)

        # now add filesystems on this pool
        for fs_info in stratis_info.filesystems.values():
            if fs_info.pool_uuid != pool_info.uuid:
                continue

            fs_device = self._devicetree.get_device_by_uuid(fs_info.uuid)
            if fs_device is not None:
                log.debug("stratis filesystem already added %s", fs_info.name)
                continue

            pool_device = self._devicetree.get_device_by_uuid(fs_info.pool_uuid)
            if not pool_device:
                log.info("stratis pool %s has not been added yet", fs_info.pool_name)
                return

            fs_device = StratisFilesystemDevice(fs_info.name, parents=[pool_device],
                                                uuid=fs_info.uuid, size=STRATIS_FS_SIZE,
                                                exists=True)
            self._devicetree._add_device(fs_device)

            # do format handling now
            udev_info = udev.get_device(fs_device.sysfs_path)
            if not udev_info:
                return

            self._devicetree.handle_format(udev_info, fs_device)
            fs_device.original_format = copy.deepcopy(fs_device.format)

    def run(self):
        log_method_call(self, pv=self.device.name)
        super(StratisFormatPopulator, self).run()

        if not self.device.format.locked_pool:
            self._add_pool_device()


class StratisXFSFormatPopulator(FormatPopulator):
    priority = 100
    _type_specifier = "stratis xfs"

    @classmethod
    def match(cls, data, device):  # pylint: disable=arguments-differ,unused-argument
        """ Return True if this helper is appropriate for the given device.

            :param :class:`pyudev.Device` data: udev data describing a device
            :param device: device instance corresponding to the udev data
            :type device: :class:`~.devices.StorageDevice`
            :returns: whether this class is appropriate for the specified device
            :rtype: bool
        """
        if device.type == "stratis filesystem" and udev.device_get_format(data) == "xfs":
            return True

        return False

    def run(self):
        """ Create a format instance and associate it with the device instance. """
        kwargs = self._get_kwargs()
        kwargs["pool_uuid"] = self.device.pool.uuid
        log.info("type detected on '%s' is '%s'", self.device.name, self.type_spec)
        self.device.format = get_format(self.type_spec, **kwargs)



#
# nvme.py - NVMe class
#
# Copyright (C) 2022  Red Hat, Inc.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

import os

from . import errors

import gi
gi.require_version("BlockDev", "3.0")

from gi.repository import BlockDev as blockdev

import logging
log = logging.getLogger("blivet")

ETC_NVME_PATH = "/etc/nvme/"
HOSTNQN_FILE = ETC_NVME_PATH + "hostnqn"
HOSTID_FILE = ETC_NVME_PATH + "hostid"


class NVMe(object):
    """ NVMe utility class.

        .. warning::
            Since this is a singleton class, calling deepcopy() on the instance
            just returns ``self`` with no copy being created.
    """

    def __init__(self):
        self.started = False
        self._hostnqn = None
        self._hostid = None

    # So that users can write nvme() to get the singleton instance
    def __call__(self):
        return self

    def __deepcopy__(self, memo_dict):  # pylint: disable=unused-argument
        return self

    def startup(self):
        if self.started:
            return

        self._hostnqn = blockdev.nvme_get_host_nqn()
        self._hostid = blockdev.nvme_get_host_id()
        if not self._hostnqn:
            self._hostnqn = blockdev.nvme_generate_host_nqn()
        if not self._hostnqn:
            raise errors.NVMeError("Failed to generate HostNQN")
        if not self._hostid:
            if 'uuid:' not in self._hostnqn:
                raise errors.NVMeError("Missing UUID part in the HostNQN string '%s'" % self._hostnqn)
            # derive HostID from HostNQN's UUID part
            self._hostid = self._hostnqn.split('uuid:')[1]

        # do not overwrite existing files, taken e.g. from initramfs
        self.write("/", overwrite=False)

        self.started = True

    def write(self, root, overwrite=True):  # pylint: disable=unused-argument
        # write down the hostnqn and hostid files
        p = root + ETC_NVME_PATH
        if not os.path.isdir(p):
            os.makedirs(p, 0o755)
        p = root + HOSTNQN_FILE
        if overwrite or not os.path.isfile(p):
            with open(p, "w") as f:
                f.write(self._hostnqn)
                f.write("\n")
        p = root + HOSTID_FILE
        if overwrite or not os.path.isfile(p):
            with open(p, "w") as f:
                f.write(self._hostid)
                f.write("\n")


# Create nvme singleton
nvme = NVMe()
""" An instance of :class:`NVMe` """



#
# mdraid.py
# mdraid functions
#
# Copyright (C) 2009-2014  Red Hat, Inc.  All rights reserved.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
# Author(s): Dave Lehman <dlehman@redhat.com>
#

from ..size import Size
from . import raid
from ..tasks import availability

import logging
log = logging.getLogger("blivet")

# these defaults were determined empirically
MD_SUPERBLOCK_SIZE = Size("2 MiB")
MD_CHUNK_SIZE = Size("512 KiB")

safe_name_characters = "0-9a-zA-Z._-"


class MDRaidLevels(raid.RAIDLevels):

    @classmethod
    def is_raid_level(cls, level):
        return super(MDRaidLevels, cls).is_raid_level(level) and \
            hasattr(level, 'get_max_spares') and \
            hasattr(level, 'get_space') and \
            hasattr(level, 'get_recommended_stride') and \
            hasattr(level, 'get_size')


raid_levels = MDRaidLevels(["raid0", "raid1", "raid4", "raid5", "raid6", "raid10", "linear"])

EXTERNAL_DEPENDENCIES = [availability.BLOCKDEV_MDRAID_PLUGIN]



# fssync.py
# Filesystem syncing classes.
#
# Copyright (C) 2014  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Anne Mulhern <amulhern@redhat.com>

import abc

from six import add_metaclass

from ..errors import FSError
from .. import util

from . import availability
from . import fstask
from . import task


@add_metaclass(abc.ABCMeta)
class FSSync(task.BasicApplication, fstask.FSTask):

    """ An abstract class that represents syncing a filesystem. """

    description = "filesystem syncing"

    @abc.abstractmethod
    def do_task(self):  # pylint: disable=arguments-differ
        raise NotImplementedError()


class XFSSync(FSSync):

    """ Sync application for XFS. """

    ext = availability.XFSFREEZE_APP

    def _get_mountpoint(self, root=None):
        mountpoint = self.fs.system_mountpoint
        if root is not None and root.replace('/', ''):
            if mountpoint == root:
                mountpoint = '/'
            else:
                mountpoint = mountpoint[len(root):]

        return mountpoint

    def _freeze_command(self, root=None):
        return [str(self.ext), "-f", self._get_mountpoint(root=root)]

    def _unfreeze_command(self, root=None):
        return [str(self.ext), "-u", self._get_mountpoint(root=root)]

    def do_task(self, root="/"):
        # pylint: disable=arguments-differ
        error_msgs = self.availability_errors
        if error_msgs:
            raise FSError("\n".join(error_msgs))

        error_msg = None
        try:
            rc = util.run_program(self._freeze_command(root=root), root=root)
        except OSError as e:
            error_msg = "failed to sync filesytem: %s" % e
        error_msg = error_msg or rc

        try:
            rc = util.run_program(self._unfreeze_command(root=root), root=root)
        except OSError as e:
            error_msg = error_msg or "failed to sync filesystem: %s" % e
        error_msg = error_msg or rc

        if error_msg:
            raise FSError(error_msg)


class UnimplementedFSSync(fstask.UnimplementedFSTask):
    pass



# events/changes.py
# Per-thread change accounting data structures.
#
# Copyright (C) 2016  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU Lesser General Public License v.2, or (at your option) any later
# version. This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY expressed or implied, including the implied
# warranties of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See
# the GNU Lesser General Public License for more details.  You should have
# received a copy of the GNU Lesser General Public License along with this
# program; if not, write to the Free Software Foundation, Inc., 51 Franklin
# Street, Fifth Floor, Boston, MA 02110-1301, USA.  Any Red Hat trademarks
# that are incorporated in the source code or documentation are not subject
# to the GNU Lesser General Public License and may only be used or
# replicated with the express permission of Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

from threading import local

from ..callbacks import callbacks
from .. import util


data = local()
""" Thread-local data for event handler threads. """

data.changes = list()  # Saves checking if the list exists in the main thread.


""" Accounting for changes made in response to an event.

    We want to end up with a single, ordered list containing every change
    that was made in response to a specific event.
"""
_DeviceAdded = util.default_namedtuple("DeviceAdded", ["device"])
_DeviceRemoved = util.default_namedtuple("DeviceRemoved", ["device"])
_ActionCanceled = util.default_namedtuple("ActionCanceled", ["action"])
_ParentAdded = util.default_namedtuple("ListItemAdded", ["device", "item"])
_ParentRemoved = util.default_namedtuple("ListItemRemoved", ["device", "item"])
_AttributeChanged = util.default_namedtuple("AttrChanged", ["device", ("fmt", None), "attr", "old", "new"])


class DeviceAdded(_DeviceAdded):
    def __str__(self):
        return "add device '%s'" % str(self.device)


class DeviceRemoved(_DeviceRemoved):
    def __str__(self):
        return "remove device %s" % self.device.name


class ActionCanceled(_ActionCanceled):
    def __str__(self):
        return "cancel action '%s'" % str(self.action)


class ParentAdded(_ParentAdded):
    def __str__(self):
        return "add parent %s to %s" % (self.item.name, self.device.name)


class ParentRemoved(_ParentRemoved):
    def __str__(self):
        return "remove parent %s from %s" % (self.item.name, self.device.name)


class AttributeChanged(_AttributeChanged):
    def __str__(self):
        return "change attribute %s of %s%s from '%s' to '%s'" % (self.attr,
                                                                  self.device.name,
                                                                  " format" if self.fmt else "",
                                                                  str(self.old),
                                                                  str(self.new))


def record_change(change):
    if not hasattr(data, "changes"):
        data.changes = list()

    data.changes.append(change)


def device_added_cb(device):
    record_change(DeviceAdded(device))


def device_removed_cb(device):
    record_change(DeviceRemoved(device))


def action_removed_cb(action):
    record_change(ActionCanceled(action))


def parent_added_cb(device, parent):
    record_change(ParentAdded(parent, device))


def parent_removed_cb(device, parent):
    record_change(ParentRemoved(parent, device))


def attribute_changed_cb(device, attr, old, new, fmt=None):
    record_change(AttributeChanged(device, fmt, attr, old, new))


def _control_callbacks(meth):
    getattr(callbacks.device_added, meth)(device_added_cb)
    getattr(callbacks.device_removed, meth)(device_removed_cb)
    getattr(callbacks.action_removed, meth)(action_removed_cb)
    getattr(callbacks.parent_added, meth)(parent_added_cb)
    getattr(callbacks.parent_removed, meth)(parent_removed_cb)
    getattr(callbacks.attribute_changed, meth)(attribute_changed_cb)


def enable_callbacks():
    _control_callbacks("add")


def disable_callbacks():
    _control_callbacks("remove")



# coding=utf-8
# pvtask.py
# Tasks for a LVMPV format.
#
# Copyright (C) 2016  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Vojtěch Trefný <vtrefny@redhat.com>

import gi
gi.require_version("BlockDev", "3.0")

from gi.repository import BlockDev as blockdev

from ..errors import PhysicalVolumeError
from ..size import Size, B

from . import availability
from . import task
from . import dfresize


class PVSize(task.BasicApplication):
    """ Obtain information about the size of a LVMPV format. """

    ext = availability.BLOCKDEV_LVM_PLUGIN

    description = "size of a LVMPV format"

    def __init__(self, a_pv):
        """ Initializer.

            :param :class:`~.formats.lvmpv.LVMPhysicalVolume` a_pv: a LVMPV format object
        """
        self.pv = a_pv

    def do_task(self):  # pylint: disable=arguments-differ
        """ Returns the size of the LVMPV format.

            :returns: the size of the LVMPV format
            :rtype: :class:`~.size.Size`
            :raises :class:`~.errors.PhysicalVolumeError`: if size cannot be obtained
        """

        try:
            pv_info = blockdev.lvm.pvinfo(self.pv.device)
            pv_size = pv_info.pv_size
        except blockdev.LVMError as e:
            raise PhysicalVolumeError(e)

        return Size(pv_size)


class PVResize(task.BasicApplication, dfresize.DFResizeTask):
    """ Handle resize of the LVMPV format. """

    description = "resize the LVMPV format"

    ext = availability.BLOCKDEV_LVM_PLUGIN
    unit = B

    def __init__(self, a_pv):
        """ Initializer.

            :param :class:`~.formats.lvmpv.LVMPhysicalVolume` a_pv: a LVMPV format object
        """
        self.pv = a_pv

    def do_task(self):  # pylint: disable=arguments-differ
        """ Resizes the LVMPV format. """
        try:
            blockdev.lvm.pvresize(self.pv.device, self.pv.target_size.convert_to(self.unit))
        except blockdev.LVMError as e:
            raise PhysicalVolumeError(e)



# errors.py
# Exception classes for anaconda's storage configuration module.
#
# Copyright (C) 2009  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Dave Lehman <dlehman@redhat.com>
#

from .i18n import N_


class StorageError(Exception):

    def __init__(self, *args, **kwargs):
        self.hardware_fault = kwargs.pop("hardware_fault", False)
        super(StorageError, self).__init__(*args, **kwargs)


class NoDisksError(StorageError):
    pass

# Device


class DeviceError(StorageError):
    pass


class DeviceCreateError(DeviceError):
    pass


class DeviceDestroyError(DeviceError):
    pass


class DeviceResizeError(DeviceError):
    pass


class DeviceSetupError(DeviceError):
    pass


class DeviceTeardownError(DeviceError):
    pass


class DeviceUserDeniedFormatError(DeviceError):
    pass


class InconsistentPVSectorSize(DeviceError, ValueError):
    pass

# DeviceFormat


class DeviceFormatError(StorageError):
    pass


class FormatCreateError(DeviceFormatError):
    pass


class FormatDestroyError(DeviceFormatError):
    pass


class FormatSetupError(DeviceFormatError):
    pass


class FormatTeardownError(DeviceFormatError):
    pass


class FormatResizeError(DeviceFormatError):

    def __init__(self, message, details):
        DeviceFormatError.__init__(self, message)
        self.details = details


class DMRaidMemberError(DeviceFormatError):
    pass


class MultipathMemberError(DeviceFormatError):
    pass


class FSError(DeviceFormatError):
    pass


class FSWriteLabelError(FSError):
    pass


class FSWriteUUIDError(FSError):
    pass


class FSReadLabelError(FSError):
    pass


class FSResizeError(FSError):

    def __init__(self, message, details):
        FSError.__init__(self, message)
        self.details = details


class LUKSError(DeviceFormatError):
    pass


class IntegrityError(LUKSError):
    pass


class MDMemberError(DeviceFormatError):
    pass


class PhysicalVolumeError(DeviceFormatError):
    pass


class SinglePhysicalVolumeError(DeviceFormatError):
    pass


class SwapSpaceError(DeviceFormatError):
    pass


class DiskLabelError(DeviceFormatError):
    pass


class InvalidDiskLabelError(DiskLabelError):
    pass


class DiskLabelCommitError(DiskLabelError):
    pass


class AlignmentError(DiskLabelError):
    pass

# devicelibs


class RaidError(StorageError):
    pass


class DMError(StorageError):
    pass


class MPathError(StorageError):
    pass


class BTRFSError(StorageError):
    pass


class BTRFSValueError(BTRFSError, ValueError):
    pass


class StratisError(StorageError):
    pass


class LoopError(StorageError):
    pass


class MDRaidError(StorageError):
    pass


class LVMError(StorageError):
    pass

# DeviceTree


class DeviceTreeError(StorageError):
    pass


class NoParentsError(DeviceTreeError):
    pass


class DeviceNotFoundError(StorageError):
    pass


class UnusableConfigurationError(StorageError):

    """ User has an unusable initial storage configuration. """
    suggestion = ""

    def __init__(self, message, dev_name=None):
        super(UnusableConfigurationError, self).__init__(message)
        self.dev_name = dev_name


class DuplicateUUIDError(UnusableConfigurationError, ValueError):
    suggestion = N_("This is usually caused by cloning the device image resulting "
                    "in duplication of the UUID value which should be unique. "
                    "In that case you can either disconnect one of the devices or "
                    "reformat it.")


class DiskLabelScanError(UnusableConfigurationError):
    suggestion = N_("For some reason we were unable to locate a disklabel on a "
                    "disk that the kernel is reporting partitions on. It is "
                    "unclear what the exact problem is. Please file a bug at "
                    "http://bugzilla.redhat.com")


class CorruptGPTError(UnusableConfigurationError):
    suggestion = N_("Either restore the disklabel to a completely working "
                    "state or remove it completely.\n"
                    "Hint: parted can restore it or wipefs can remove it.")


class DuplicateVGError(UnusableConfigurationError):
    suggestion = N_("Rename one of the volume groups so the names are "
                    "distinct.\n"
                    "Hint 1: vgrename accepts UUID in place of the old name.\n"
                    "Hint 2: You can get the VG UUIDs by running "
                    "'pvs -o +vg_uuid'.")


class InvalidMultideviceSelection(UnusableConfigurationError):
    suggestion = N_("All parent devices must be selected when choosing exclusive "
                    "or ignored disks for a multipath or firmware RAID device.")


# DeviceAction


class DeviceActionError(StorageError):
    pass

# partitioning


class PartitioningError(StorageError):
    pass


class NotEnoughFreeSpaceError(StorageError):
    pass

# udev


class UdevError(StorageError):
    pass

# fstab


class UnrecognizedFSTabEntryError(StorageError):
    pass


class FSTabTypeMismatchError(StorageError):
    pass

# factories


class DeviceFactoryError(StorageError):
    pass


class AvailabilityError(StorageError):

    """ Raised if problem determining availability of external resource. """


class EventManagerError(StorageError):
    pass


class EventParamError(StorageError):
    pass

# external dependencies


class DependencyError(StorageError):
    """Raised when an external dependency is missing or not available"""


class EventHandlingError(StorageError):
    pass


class ThreadError(StorageError):
    """ An error occurred in a non-main thread. """

# other


class FCoEError(StorageError, OSError):
    pass


class ISCSIError(StorageError, OSError):
    pass


class NVMeError(StorageError, OSError):
    pass


class GPTVolUUIDError(StorageError):
    pass



# partitioning.py
# Disk partitioning functions.
#
# Copyright (C) 2009, 2010, 2011, 2012, 2013  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Dave Lehman <dlehman@redhat.com>
#

from operator import gt, lt
from decimal import Decimal
import functools

import gi
gi.require_version("BlockDev", "3.0")

import parted
import _ped

from .errors import DeviceError, PartitioningError, AlignmentError
from .flags import flags
from .devices import Device, PartitionDevice, device_path_to_name
from .size import Size
from .i18n import _
from .util import stringize, unicodeize, compare

import logging
log = logging.getLogger("blivet")


def partition_compare(part1, part2):
    """ More specifically defined partitions come first.

        < 1 => x < y
          0 => x == y
        > 1 => x > y

        :param part1: the first partition
        :type part1: :class:`devices.PartitionDevice`
        :param part2: the other partition
        :type part2: :class:`devices.PartitionDevice`
        :return: see above
        :rtype: int
    """
    ret = 0

    # start sector overrides all other sorting factors
    part1_start = part1.req_start_sector
    part2_start = part2.req_start_sector
    if part1_start is not None and part2_start is None:
        return -1
    elif part1_start is None and part2_start is not None:
        return 1
    elif part1_start is not None and part2_start is not None:
        return compare(part1_start, part2_start)

    if part1.weight:
        ret -= part1.weight

    if part2.weight:
        ret += part2.weight

    # more specific disk specs to the front of the list
    # req_disks being empty is equivalent to it being an infinitely long list
    if part1.req_disks and not part2.req_disks:
        ret -= 500
    elif not part1.req_disks and part2.req_disks:
        ret += 500
    else:
        ret += compare(len(part1.req_disks), len(part2.req_disks)) * 500

    # primary-only to the front of the list
    ret -= compare(part1.req_primary, part2.req_primary) * 200

    # fixed size requests to the front
    ret += compare(part1.req_grow, part2.req_grow) * 100

    # larger requests go to the front of the list
    ret -= compare(part1.req_base_size, part2.req_base_size) * 50

    # potentially larger growable requests go to the front
    if part1.req_grow and part2.req_grow:
        if not part1.req_max_size and part2.req_max_size:
            ret -= 25
        elif part1.req_max_size and not part2.req_max_size:
            ret += 25
        else:
            ret -= compare(part1.req_max_size, part2.req_max_size) * 25

    # give a little bump based on mountpoint
    if hasattr(part1.format, "mountpoint") and \
       hasattr(part2.format, "mountpoint"):
        ret += compare(part1.format.mountpoint, part2.format.mountpoint) * 10

    if ret > 0:
        ret = 1
    elif ret < 0:
        ret = -1

    return ret


_partition_compare_key = functools.cmp_to_key(partition_compare)


def get_next_partition_type(disk, no_primary=None):
    """ Return the type of partition to create next on a disk.

        Return a parted partition type value representing the type of the
        next partition we will create on this disk.

        If there is only one free primary partition and we can create an
        extended partition, we do that.

        If there are free primary slots and an extended partition we will
        recommend creating a primary partition. This can be overridden
        with the keyword argument no_primary.

        :param disk: the disk from which a partition may be allocated
        :type disk: :class:`parted.Disk`
        :keyword no_primary: refuse to return :const:`parted.PARTITION_NORMAL`
        :returns: the chosen partition type
        :rtype: a parted PARTITION_* constant
    """
    part_type = None
    extended = disk.getExtendedPartition()
    supports_extended = disk.supportsFeature(parted.DISK_TYPE_EXTENDED)
    primary_count = disk.primaryPartitionCount

    if primary_count < disk.maxPrimaryPartitionCount:
        if primary_count == disk.maxPrimaryPartitionCount - 1:
            # can we make an extended partition? now's our chance.
            if not extended and supports_extended:
                part_type = parted.PARTITION_EXTENDED
            elif not extended:
                # extended partitions not supported. primary or nothing.
                if not no_primary:
                    part_type = parted.PARTITION_NORMAL
            else:
                # there is an extended and a free primary
                if not no_primary:
                    part_type = parted.PARTITION_NORMAL
                else:
                    # we have an extended, so use it.
                    part_type = parted.PARTITION_LOGICAL
        else:
            # there are two or more primary slots left. use one unless we're
            # not supposed to make primaries.
            if not no_primary:
                part_type = parted.PARTITION_NORMAL
            elif extended:
                part_type = parted.PARTITION_LOGICAL
    elif extended:
        part_type = parted.PARTITION_LOGICAL

    return part_type


def get_best_free_space_region(disk, part_type, req_size, start=None,
                               boot=None, best_free=None, grow=None,
                               alignment=None):
    """ Return the "best" free region on the specified disk.

        For non-boot partitions, we return the largest free region on the
        disk. For boot partitions, we return the first region that is
        large enough to hold the partition.

        Partition type (parted's PARTITION_NORMAL, PARTITION_LOGICAL) is
        taken into account when locating a suitable free region.

        For locating the best region from among several disks, the keyword
        argument best_free allows the specification of a current "best"
        free region with which to compare the best from this disk. The
        overall best region is returned.

        :param disk: the disk
        :type disk: :class:`parted.Disk`
        :param part_type: the type of partition we want to allocate
        :type part_type: one of parted's PARTITION_* constants
        :param req_size: the requested size of the partition in MiB
        :type req_size: :class:`~.size.Size`
        :keyword start: requested start sector for the partition
        :type start: int
        :keyword boot: whether this will be a bootable partition
        :type boot: bool
        :keyword best_free: current best free region for this partition
        :type best_free: :class:`parted.Geometry`
        :keyword grow: indicates whether this is a growable request
        :type grow: bool
        :keyword alignment: disk alignment requirements
        :type alignment: :class:`parted.Alignment`

    """
    log.debug("get_best_free_space_region: disk=%s part_type=%d req_size=%s "
              "boot=%s best=%s grow=%s start=%s",
              disk.device.path, part_type, req_size, boot, best_free, grow,
              start)
    extended = disk.getExtendedPartition()
    alignment = alignment or parted.Alignment(offset=0, grainSize=1)

    for free_geom in disk.getFreeSpaceRegions():
        # align the start sector of the free region since we will be aligning
        # the start sector of the partition
        if start is not None and \
           not alignment.isAligned(free_geom, free_geom.start):
            log.debug("aligning start sector of region %d-%d", free_geom.start,
                      free_geom.end)
            try:
                aligned_start = alignment.alignUp(free_geom, free_geom.start)
            except ArithmeticError:
                aligned_start = None
            else:
                # parted tends to align down when it cannot align up
                if aligned_start < free_geom.start:
                    aligned_start = None

            if aligned_start is None:
                log.debug("failed to align start sector -- skipping region")
                continue

            free_geom = parted.Geometry(device=free_geom.device,
                                        start=aligned_start,
                                        end=free_geom.end)

        log.debug("checking %d-%d (%s)", free_geom.start, free_geom.end,
                  Size(free_geom.getLength(unit="B")))
        if start is not None and not free_geom.containsSector(start):
            log.debug("free region does not contain requested start sector")
            continue

        if extended:
            in_extended = extended.geometry.contains(free_geom)
            if ((in_extended and part_type == parted.PARTITION_NORMAL) or
                    (not in_extended and part_type == parted.PARTITION_LOGICAL)):
                log.debug("free region not suitable for request")
                continue

        if free_geom.start > disk.maxPartitionStartSector:
            log.debug("free range start sector beyond max for new partitions")
            continue

        if boot:
            max_boot = Size("2 TiB")
            free_start = Size(free_geom.start * disk.device.sectorSize)
            req_end = free_start + req_size
            if req_end > max_boot:
                log.debug("free range position would place boot req above %s",
                          max_boot)
                continue

        log.debug("current free range is %d-%d (%s)", free_geom.start,
                  free_geom.end,
                  Size(free_geom.getLength(unit="B")))
        free_size = Size(free_geom.getLength(unit="B"))

        # For boot partitions, we want the first suitable region we find.
        # For growable or extended partitions, we want the largest possible
        # free region.
        # For all others, we want the smallest suitable free region.
        if grow or part_type == parted.PARTITION_EXTENDED:
            op = gt
        else:
            op = lt
        if req_size <= free_size:
            if not best_free or op(free_geom.length, best_free.length):
                best_free = free_geom

                if boot:
                    # if this is a bootable partition we want to
                    # use the first freespace region large enough
                    # to satisfy the request
                    break

    return best_free


def sectors_to_size(sectors, sector_size):
    """ Convert length in sectors to size.

        :param sectors: sector count
        :type sectors: int
        :param sector_size: sector size
        :type sector_size: :class:`~.size.Size`
        :returns: the size
        :rtype: :class:`~.size.Size`
    """
    return Size(sectors * sector_size)


def size_to_sectors(size, sector_size):
    """ Convert size to length in sectors.

        :param size: size
        :type size: :class:`~.size.Size`
        :param sector_size: sector size in bytes
        :type sector_size: :class:`~.size.Size`
        :returns: sector count
        :rtype: int
    """
    return int(size // sector_size)


def remove_new_partitions(disks, remove, all_partitions):
    """ Remove newly added partitions from disks.

        Remove all non-existent partitions from the disks in blivet's model.

        :param: disks: list of partitioned disks
        :type disks: list of :class:`~.devices.StorageDevice`
        :param remove: list of partitions to remove
        :type remove: list of :class:`~.devices.PartitionDevice`
        :param all_partitions: list of all partitions on the disks
        :type all_partitions: list of :class:`~.devices.PartitionDevice`
        :returns: None
        :rtype: NoneType
    """
    log.debug("removing all non-preexisting partitions %s from disk(s) %s",
              ["%s(id %d)" % (p.name, p.id) for p in remove],
              [d.name for d in disks])

    removed_logical = []
    for part in remove:
        if part.parted_partition and part.disk in disks:
            if part.exists:
                # we're only removing partitions that don't physically exist
                continue

            if part.is_extended:
                # these get removed last
                continue

            if part.is_logical:
                removed_logical.append(part)
            part.disk.format.parted_disk.removePartition(part.parted_partition)
            part.parted_partition = None
            part.disk = None

    def _remove_extended(disk, extended):
        """ We may want to remove extended partition from the disk too.
            This should happen if we don't have the PartitionDevice object
            or in installer_mode after we've removed all logical paritions.
        """

        if extended and not disk.format.logical_partitions:
            if extended not in (p.parted_partition for p in all_partitions):
                # extended partition is not in all_partitions -> remove it
                return True
            else:
                if flags.keep_empty_ext_partitions:
                    return False
                else:
                    if any(l.disk == extended.disk for l in removed_logical):
                        # we removed all logical paritions from this extended
                        # so we no longer need this one
                        return True
                    else:
                        return False

    for disk in disks:
        # remove empty extended so it doesn't interfere
        extended = disk.format.extended_partition
        if _remove_extended(disk, extended):
            log.debug("removing empty extended partition from %s", disk.name)
            disk.format.parted_disk.removePartition(extended)


def add_partition(disklabel, free, part_type, size, start=None, end=None):
    """ Add a new partition to a disk.

        :param disklabel: the disklabel to add the partition to
        :type disklabel: :class:`~.formats.DiskLabel`
        :param free: the free region in which to place the new partition
        :type free: :class:`parted.Geometry`
        :param part_type: the partition type
        :type part_type: a parted.PARTITION_* constant
        :param size: size of the new partition
        :type size: :class:`~.size.Size`
        :keyword start: starting sector for the partition
        :type start: int
        :keyword end: ending sector for the partition
        :type end: int
        :raises: :class:`~.errors.PartitioningError`
        :returns: the newly added partitions
        :rtype: :class:`parted.Partition`

        .. note::

            The new partition will be aligned using the kernel-provided optimal
            alignment unless a start sector is provided.

    """
    # get alignment information based on disklabel, device, and partition size
    if start is None:
        if size is None:
            # implicit request for extended partition (will use full free area)
            _size = sectors_to_size(free.length, disklabel.sector_size)
        else:
            _size = size

        try:
            alignment = disklabel.get_alignment(size=_size)
        except AlignmentError:
            alignment = disklabel.get_minimal_alignment()

        end_alignment = disklabel.get_end_alignment(alignment=alignment)
    else:
        alignment = parted.Alignment(grainSize=1, offset=0)
        end_alignment = parted.Alignment(grainSize=1, offset=-1)
    log.debug("using alignment: %s", alignment)

    sector_size = Size(disklabel.sector_size)
    if start is not None:
        if end is None:
            end = start + size_to_sectors(size, sector_size) - 1
    else:
        start = free.start

        if not alignment.isAligned(free, start):
            start = alignment.alignNearest(free, start)

        if disklabel.label_type == "sun" and start == 0:
            start = alignment.alignUp(free, start)

        if part_type == parted.PARTITION_LOGICAL:
            # make room for logical partition's metadata
            start += alignment.grainSize

        if start != free.start:
            log.debug("adjusted start sector from %d to %d", free.start, start)

        if part_type == parted.PARTITION_EXTENDED and not size:
            end = free.end
            length = end - start + 1
        else:
            length = size_to_sectors(size, sector_size)
            end = start + length - 1

        if not end_alignment.isAligned(free, end):
            end = end_alignment.alignUp(free, end)
            log.debug("adjusted length from %d to %d", length, end - start + 1)
            if start > end:
                raise PartitioningError(_("unable to allocate aligned partition"))

    new_geom = parted.Geometry(device=disklabel.parted_device,
                               start=start,
                               end=end)

    max_length = disklabel.parted_disk.maxPartitionLength
    if max_length and new_geom.length > max_length:
        raise PartitioningError(_("requested size exceeds maximum allowed"))

    # create the partition and add it to the disk
    partition = parted.Partition(disk=disklabel.parted_disk,
                                 type=part_type,
                                 geometry=new_geom)
    constraint = parted.Constraint(exactGeom=new_geom)

    try:
        disklabel.parted_disk.addPartition(partition=partition,
                                           constraint=constraint)
    except _ped.PartitionException as e:
        raise PartitioningError(_("failed to add partition to disk: %s") % str(e))

    return partition


def get_free_regions(disks, align=False):
    """ Return a list of free regions on the specified disks.

        :param disks: list of disks
        :type disks: list of :class:`~.devices.Disk`
        :param align: align the region length to disk grain_size
        :type align: bool
        :returns: list of free regions
        :rtype: list of :class:`parted.Geometry`

        Only free regions guaranteed to contain at least one aligned sector for
        both the start and end alignments in the
        :class:`~.formats.disklabel.DiskLabel` are returned.
    """
    free = []
    for disk in disks:
        for f in disk.format.parted_disk.getFreeSpaceRegions():
            grain_size = disk.format.alignment.grainSize
            if f.length >= grain_size:
                if align:
                    aligned_length = f.length - (f.length % grain_size)
                    log.debug("length of free region aligned from %d to %d",
                              f.length, aligned_length)
                    f.length = aligned_length
                free.append(f)

    return free


def update_extended_partitions(storage, disks):
    """ Reconcile extended partition changes with the DeviceTree.

        :param storage: the Blivet instance
        :type storage: :class:`~.Blivet`
        :param disks: list of disks
        :type disks: list of :class:`~.devices.StorageDevice`
        :returns: :const:`None`
        :rtype: NoneType
    """
    # XXX hack -- if we created any extended partitions we need to add
    #             them to the tree now
    for disk in disks:
        extended = disk.format.extended_partition
        if not extended:
            # remove any obsolete extended partitions
            for part in storage.partitions:
                if part.disk == disk and part.is_extended:
                    if part.exists:
                        storage.destroy_device(part)
                    else:
                        storage.devicetree._remove_device(part, modparent=False)
            continue

        extended_name = device_path_to_name(extended.getDeviceNodeName())
        device = storage.devicetree.get_device_by_name(extended_name)
        if device:
            if not device.exists:
                # created by us, update parted_partition
                device.parted_partition = extended

        # remove any obsolete extended partitions
        for part in storage.partitions:
            if part.disk == disk and part.is_extended and \
               part.parted_partition not in disk.format.partitions:
                if part.exists:
                    storage.destroy_device(part)
                else:
                    storage.devicetree._remove_device(part, modparent=False)

        if device:
            continue

        # This is a little odd because normally instantiating a partition
        # that does not exist means leaving self.parents empty and instead
        # populating self.req_disks. In this case, we need to skip past
        # that since this partition is already defined.
        device = PartitionDevice(extended_name, parents=disk)
        device.parents = [disk]
        device.parted_partition = extended
        # just add the device for now -- we'll handle actions at the last
        # moment to simplify things
        storage.devicetree._add_device(device)


def do_partitioning(storage, boot_disk=None):
    """ Allocate and grow partitions.

        When this function returns without error, all PartitionDevice
        instances must have their parents set to the disk they are
        allocated on, and their parted_partition attribute set to the
        appropriate parted.Partition instance from their containing
        disk. All req_xxxx attributes must be unchanged.

        :param storage: Blivet instance
        :type storage: :class:`~.Blivet`
        :param boot_disk: optional parameter, the disk the bootloader is on
        :type boot_device: :class:`~.devices.StorageDevice`
        :raises: :class:`~.errors.PartitioningError`
        :returns: :const:`None`
    """
    disks = [d for d in storage.partitioned if d.format.supported and not d.protected]
    for disk in disks:
        try:
            disk.setup()
        except DeviceError as e:
            log.error("failed to set up disk %s: %s", disk.name, e)
            raise PartitioningError(_("disk %s inaccessible") % disk.name)

    # Remove any extended partition that does not have an action associated.
    #
    # XXX This does not remove the extended from the parted.Disk, but it should
    #     cause remove_new_partitions to remove it since there will no longer be
    #     a PartitionDevice for it.
    for partition in storage.partitions:
        if not partition.exists and partition.is_extended and \
           not storage.devicetree.actions.find(device=partition, action_type="create"):
            storage.devicetree._remove_device(partition, modparent=False, force=True)

    partitions = storage.partitions[:]

    if boot_disk:
        # set the boot flag on boot device
        for part in storage.partitions:
            part.req_bootable = False
            if not part.exists:
                # start over with flexible-size requests
                part.req_size = part.req_base_size

        boot_device = storage.mountpoints.get("/boot", storage.mountpoints.get("/"))
        try:
            boot_device.req_bootable = True
        except AttributeError:
            # there's no stage2 device. hopefully it's temporary.
            pass

    remove_new_partitions(disks, partitions, partitions)
    free = get_free_regions(disks)
    try:
        allocate_partitions(storage, disks, partitions, free, boot_disk=boot_disk)
        grow_partitions(disks, partitions, free, size_sets=storage.size_sets)
    except Exception:  # pylint: disable=try-except-raise
        raise
    else:
        # Mark all growable requests as no longer growable.
        for partition in storage.partitions:
            log.debug("fixing size of %s", partition)
            partition.req_grow = False
            partition.req_base_size = partition.size
            partition.req_size = partition.size
    finally:
        # these are only valid for one allocation run
        storage.size_sets = []

        # The number and thus the name of partitions may have changed now,
        # allocate_partitions() takes care of this for new partitions, but not
        # for pre-existing ones, so we update the name of all partitions here
        for part in storage.partitions:
            # leave extended partitions as-is -- we'll handle them separately
            if part.is_extended:
                continue
            part.update_name()

        update_extended_partitions(storage, disks)

        for part in [p for p in storage.partitions if not p.exists]:
            problem = part.check_size()
            if problem < 0:
                raise PartitioningError(_("partition is too small for %(format)s formatting "
                                          "(allowable size is %(min_size)s to %(max_size)s)")
                                        % {"format": part.format.name, "min_size": part.format.min_size,
                                            "max_size": part.format.max_size})
            elif problem > 0:
                raise PartitioningError(_("partition is too large for %(format)s formatting "
                                          "(allowable size is %(min_size)s to %(max_size)s)")
                                        % {"format": part.format.name, "min_size": part.format.min_size,
                                            "max_size": part.format.max_size})


def align_size_for_disklabel(size, disklabel):
    # Align the base size to the disk's grain size.
    try:
        alignment = disklabel.get_alignment(size=size)
    except AlignmentError:
        alignment = disklabel.get_minimal_alignment()

    grain_size = Size(alignment.grainSize)
    grains, rem = divmod(size, grain_size)
    return (grains * grain_size) + (grain_size if rem else Size(0))


def resolve_disk_tags(disks, tags):
    """Resolve disk tags to a disk list.

        :param disks: available disks
        :type disks: list of :class:`~.devices.StorageDevice`
        :param tags: tags to select disks based on
        :type tags: list of str

        If tags contains multiple values it is interpeted as
        "include disks containing *any* of these tags".
    """
    return [disk for disk in disks if any(tag in disk.tags for tag in tags)]


def allocate_partitions(storage, disks, partitions, freespace, boot_disk=None):
    """ Allocate partitions based on requested features.

        :param storage: a Blivet instance
        :type storage: :class:`~.Blivet`
        :param disks: list of usable disks
        :type disks: list of :class:`~.devices.StorageDevice`
        :param partitions: list of partitions
        :type partitions: list of :class:`~.devices.PartitionDevice`
        :param freespace: list of free regions on disks
        :type freespace: list of :class:`parted.Geometry`
        :param boot_disk: optional parameter, the disk the bootloader is on
        :type boot_device: :class:`~.devices.StorageDevice`
        :raises: :class:`~.errors.PartitioningError`
        :returns: :const:`None`

        Non-existing partitions are sorted according to their requested
        attributes, and then allocated.

        The basic approach to sorting is that the more specifically-
        defined a request is, the earlier it will be allocated. See
        :func:`partitionCompare` for details of the sorting criteria.

        The :class:`~.devices.PartitionDevice` instances will have their name
        and parents attributes set once they have been allocated.
    """
    log.debug("allocate_partitions: disks=%s ; partitions=%s",
              [d.name for d in disks],
              ["%s(id %d)" % (p.name, p.id) for p in partitions])

    new_partitions = [p for p in partitions if not p.exists]
    new_partitions.sort(key=_partition_compare_key)

    # the following dicts all use device path strings as keys
    disklabels = {}     # DiskLabel instances for each disk
    all_disks = {}      # StorageDevice for each disk
    for disk in disks:
        if disk.path not in disklabels.keys():
            disklabels[disk.path] = disk.format
            all_disks[disk.path] = disk

    remove_new_partitions(disks, new_partitions, partitions)

    for _part in new_partitions:
        if _part.parted_partition and _part.is_extended:
            # ignore new extendeds as they are implicit requests
            continue

        # obtain the set of candidate disks
        req_disks = []
        if _part.req_disks:
            # use the requested disk set
            req_disks = _part.req_disks
        elif _part.req_disk_tags:
            req_disks = resolve_disk_tags(disks, _part.req_disk_tags)
        else:
            # no disks specified means any disk will do
            req_disks = disks

        req_disks.sort(key=storage.compare_disks_key)
        # make sure the boot disk is at the beginning of the disk list
        if boot_disk:
            for disk in req_disks[:]:
                if disk == boot_disk:
                    boot_index = req_disks.index(disk)
                    req_disks.insert(0, req_disks.pop(boot_index))

        boot = _part.weight > 1000

        log.debug("allocating partition: %s ; id: %d ; disks: %s ;\n"
                  "boot: %s ; primary: %s ; size: %s ; grow: %s ; "
                  "max_size: %s ; start: %s ; end: %s", _part.name, _part.id,
                  [d.name for d in req_disks],
                  boot, _part.req_primary,
                  _part.req_size, _part.req_grow,
                  _part.req_max_size, _part.req_start_sector,
                  _part.req_end_sector)
        free = None
        use_disk = None
        part_type = None
        growth = 0  # in sectors
        # loop through disks
        for _disk in req_disks:
            try:
                disklabel = disklabels[_disk.path]
            except KeyError:
                raise PartitioningError("Requested disk %s doesn't have a usable disklabel for partitioning" % _disk.name)
            best = None
            current_free = free
            try:
                alignment = disklabel.get_alignment(size=_part.req_size)
            except AlignmentError:
                alignment = disklabel.get_minimal_alignment()

            # for growable requests, we don't want to pass the current free
            # geometry to get_best_free_region -- this allows us to try the
            # best region from each disk and choose one based on the total
            # growth it allows
            if _part.req_grow:
                current_free = None

            log.debug("checking freespace on %s", _disk.name)

            if _part.req_start_sector is None:
                req_size = align_size_for_disklabel(_part.req_size, disklabel)
            else:
                # don't align size if start sector was specified
                req_size = _part.req_size

            if req_size != _part.req_size:
                log.debug("size %s rounded up to %s for disk %s",
                          _part.req_size, req_size, _disk.name)

            new_part_type = get_next_partition_type(disklabel.parted_disk)
            if new_part_type is None:
                # can't allocate any more partitions on this disk
                log.debug("no free partition slots on %s", _disk.name)
                continue

            if _part.req_primary and new_part_type != parted.PARTITION_NORMAL:
                if (disklabel.parted_disk.primaryPartitionCount <
                        disklabel.parted_disk.maxPrimaryPartitionCount):
                    # don't fail to create a primary if there are only three
                    # primary partitions on the disk (#505269)
                    new_part_type = parted.PARTITION_NORMAL
                else:
                    # we need a primary slot and none are free on this disk
                    log.debug("no primary slots available on %s", _disk.name)
                    continue
            elif _part.req_part_type is not None and \
                    new_part_type != _part.req_part_type:
                new_part_type = _part.req_part_type

            best = get_best_free_space_region(disklabel.parted_disk,
                                              new_part_type,
                                              req_size,
                                              start=_part.req_start_sector,
                                              best_free=current_free,
                                              boot=boot,
                                              grow=_part.req_grow,
                                              alignment=alignment)

            if best == free and not _part.req_primary and \
               new_part_type == parted.PARTITION_NORMAL:
                # see if we can do better with a logical partition
                log.debug("not enough free space for primary -- trying logical")
                new_part_type = get_next_partition_type(disklabel.parted_disk,
                                                        no_primary=True)
                if new_part_type:
                    best = get_best_free_space_region(disklabel.parted_disk,
                                                      new_part_type,
                                                      req_size,
                                                      start=_part.req_start_sector,
                                                      best_free=current_free,
                                                      boot=boot,
                                                      grow=_part.req_grow,
                                                      alignment=alignment)

            if best and free != best:
                update = True
                allocated = new_partitions[:new_partitions.index(_part) + 1]
                if any([p.req_grow for p in allocated]):
                    log.debug("evaluating growth potential for new layout")
                    new_growth = 0
                    for disk_path in disklabels.keys():
                        log.debug("calculating growth for disk %s", disk_path)
                        # Now we check, for growable requests, which of the two
                        # free regions will allow for more growth.

                        # set up chunks representing the disks' layouts
                        temp_parts = []
                        for _p in new_partitions[:new_partitions.index(_part)]:
                            if _p.disk.path == disk_path:
                                temp_parts.append(_p)

                        # add the current request to the temp disk to set up
                        # its parted_partition attribute with a base geometry
                        if disk_path == _disk.path:
                            _part_type = new_part_type
                            _free = best
                            if new_part_type == parted.PARTITION_EXTENDED and \
                               new_part_type != _part.req_part_type:
                                add_partition(disklabel, best, new_part_type,
                                              None)

                                _part_type = parted.PARTITION_LOGICAL

                                _free = get_best_free_space_region(disklabel.parted_disk,
                                                                   _part_type,
                                                                   req_size,
                                                                   start=_part.req_start_sector,
                                                                   boot=boot,
                                                                   grow=_part.req_grow,
                                                                   alignment=alignment)
                                if not _free:
                                    log.info("not enough space after adding "
                                             "extended partition for growth test")
                                    if new_part_type == parted.PARTITION_EXTENDED:
                                        e = disklabel.extended_partition
                                        disklabel.parted_disk.removePartition(e)

                                    continue

                            temp_part = None
                            try:
                                temp_part = add_partition(disklabel,
                                                          _free,
                                                          _part_type,
                                                          req_size,
                                                          _part.req_start_sector,
                                                          _part.req_end_sector)
                            except ArithmeticError:
                                log.debug("failed to allocate aligned partition "
                                          "for growth test")
                                continue

                            _part.parted_partition = temp_part
                            _part.disk = _disk
                            temp_parts.append(_part)

                        chunks = get_disk_chunks(all_disks[disk_path],
                                                 temp_parts, freespace)

                        # grow all growable requests
                        disk_growth = 0  # in sectors
                        disk_sector_size = Size(disklabels[disk_path].sector_size)
                        for chunk in chunks:
                            chunk.grow_requests()
                            # record the growth for this layout
                            new_growth += chunk.growth
                            disk_growth += chunk.growth
                            for req in chunk.requests:
                                log.debug("request %d (%s) growth: %d (%s) "
                                          "size: %s",
                                          req.device.id,
                                          req.device.name,
                                          req.growth,
                                          sectors_to_size(req.growth,
                                                          disk_sector_size),
                                          sectors_to_size(req.growth + req.base,
                                                          disk_sector_size))
                        log.debug("disk %s growth: %d (%s)",
                                  disk_path, disk_growth,
                                  sectors_to_size(disk_growth,
                                                  disk_sector_size))

                    if temp_part:
                        disklabel.parted_disk.removePartition(temp_part)
                    _part.parted_partition = None
                    _part.disk = None

                    if new_part_type == parted.PARTITION_EXTENDED:
                        e = disklabel.extended_partition
                        disklabel.parted_disk.removePartition(e)

                    log.debug("total growth: %d sectors", new_growth)

                    # update the chosen free region unless the previous
                    # choice yielded greater total growth
                    if free is not None and new_growth <= growth:
                        log.debug("keeping old free: %d <= %d", new_growth,
                                  growth)
                        update = False
                    else:
                        growth = new_growth

                if update:
                    # now we know we are choosing a new free space,
                    # so update the disk and part type
                    log.debug("updating use_disk to %s, type: %s",
                              _disk.name, new_part_type)
                    part_type = new_part_type
                    use_disk = _disk
                    log.debug("new free: %d-%d / %s", best.start,
                              best.end,
                              Size(best.getLength(unit="B")))
                    log.debug("new free allows for %d sectors of growth", growth)
                    free = best

            if free and boot:
                # if this is a bootable partition we want to
                # use the first freespace region large enough
                # to satisfy the request
                log.debug("found free space for bootable request")
                break

        if free is None:
            raise PartitioningError(_("Unable to allocate requested partition scheme."))

        _disk = use_disk
        disklabel = _disk.format
        if _part.req_start_sector is None:
            aligned_size = align_size_for_disklabel(_part.req_size, disklabel)
        else:
            # not aligned
            aligned_size = _part.req_size

        # create the extended partition if needed
        if part_type == parted.PARTITION_EXTENDED and \
           part_type != _part.req_part_type:
            log.debug("creating extended partition")
            ext = add_partition(disklabel, free, part_type, None)

            # extedned partition took all free space - make the size request smaller
            if aligned_size > (ext.geometry.length - disklabel.alignment.grainSize) * disklabel.sector_size:
                log.debug("not enough free space after creating extended "
                          "partition - shrinking the logical partition")
                aligned_size = aligned_size - (disklabel.alignment.grainSize * disklabel.sector_size)

            # now the extended partition exists, so set type to logical
            part_type = parted.PARTITION_LOGICAL

            # recalculate freespace
            log.debug("recalculating free space")
            free = get_best_free_space_region(disklabel.parted_disk,
                                              part_type,
                                              aligned_size,
                                              start=_part.req_start_sector,
                                              boot=boot,
                                              grow=_part.req_grow,
                                              alignment=disklabel.alignment)
            if not free:
                raise PartitioningError(_("not enough free space after "
                                          "creating extended partition"))

        try:
            partition = add_partition(disklabel, free, part_type, aligned_size,
                                      _part.req_start_sector, _part.req_end_sector)
        except ArithmeticError:
            raise PartitioningError(_("failed to allocate aligned partition"))

        log.debug("created partition %s of %s and added it to %s",
                  partition.getDeviceNodeName(),
                  Size(partition.getLength(unit="B")),
                  disklabel.device)

        # this one sets the name
        _part.parted_partition = partition
        _part.disk = _disk

        # parted modifies the partition in the process of adding it to
        # the disk, so we need to grab the latest version...
        _part.parted_partition = disklabel.parted_disk.getPartitionByPath(_part.path)


class Request(object):

    """ A partition request.

        Request instances are used for calculating how much to grow
        partitions.
    """

    def __init__(self, device):
        """
            :param device: the device being requested
            :type device: :class:`~.devices.StorageDevice`
        """
        self.device = device
        self.growth = 0                     # growth in sectors
        self.max_growth = 0                 # max growth in sectors
        self.done = not getattr(device, "req_grow", True)  # can we grow this
        # request more?
        self.base = 0                       # base sectors

    @property
    def reserve_request(self):
        """ Requested reserved fixed extra space for the request (in sectors) """

        # generic requests don't need any such extra space
        return 0

    @property
    def growable(self):
        """ True if this request is growable. """
        return getattr(self.device, "req_grow", True)

    @property
    def id(self):
        """ The id of the Device instance this request corresponds to. """
        return self.device.id

    def __repr__(self):
        s = ("%(type)s instance --\n"
             "id = %(id)s  name = %(name)s  growable = %(growable)s\n"
             "base = %(base)d  growth = %(growth)d  max_grow = %(max_grow)d\n"
             "done = %(done)s" %
             {"type": self.__class__.__name__, "id": self.id,
              "name": self.device.name, "growable": self.growable,
              "base": self.base, "growth": self.growth,
              "max_grow": self.max_growth, "done": self.done})
        return s


class PartitionRequest(Request):

    def __init__(self, partition):
        """
            :param partition: the partition being requested
            :type partition: :class:`~.devices.PartitionDevice`
        """
        super(PartitionRequest, self).__init__(partition)
        self.base = partition.parted_partition.geometry.length   # base sectors

        sector_size = Size(partition.parted_partition.disk.device.sectorSize)

        if partition.req_grow:
            mins = [size for size in (partition.req_max_size, partition.format.max_size)
                    if size > 0]
            req_format_max_size = min(mins) if mins else Size(0)
            limits = [l for l in (size_to_sectors(req_format_max_size, sector_size),
                                  partition.parted_partition.disk.maxPartitionLength) if l > 0]

            if limits:
                max_sectors = min(limits)
                self.max_growth = max_sectors - self.base
                if self.max_growth <= 0:
                    # max size is less than or equal to base, so we're done
                    self.done = True


class LVRequest(Request):

    def __init__(self, lv):
        """
            :param lv: the logical volume being requested
            :type lv: :class:`~.devices.LVMLogicalVolumeDevice`
        """
        super(LVRequest, self).__init__(lv)

        # Round up to nearest pe. For growable requests this will mean that
        # first growth is to fill the remainder of any unused extent.
        self.base = int(lv.vg.align(lv.size, roundup=True) // lv.vg.pe_size)

        if lv.req_grow:
            limits = [int(l // lv.vg.pe_size) for l in
                      (lv.vg.align(lv.req_max_size),
                       lv.vg.align(lv.format.max_size)) if l > Size(0)]

            if limits:
                max_units = min(limits)
                self.max_growth = max_units - self.base
                if self.max_growth <= 0:
                    # max size is less than or equal to base, so we're done
                    self.done = True

    @property
    def reserve_request(self):
        lv = self.device
        reserve = super(LVRequest, self).reserve_request
        if lv.cached:
            reserve += int(lv.vg.align(lv.cache.size, roundup=True) / lv.vg.pe_size)
        reserve += int(lv.vg.align(lv.metadata_vg_space_used, roundup=True) / lv.vg.pe_size)
        return reserve


class Chunk(object):

    """ A free region from which devices will be allocated """

    def __init__(self, length, requests=None):
        """
            :param length: the length of the chunk (units vary with subclass)
            :type length: int
            :keyword requests: list of requests to add
            :type requests: list of :class:`Request`
        """
        if not hasattr(self, "path"):
            self.path = None
        self.length = length
        self.pool = length                  # free unit count
        self.base = 0                       # sum of growable requests' base
        # sizes
        self.requests = []                  # list of Request instances
        if isinstance(requests, list):
            for req in requests:
                self.add_request(req)

        self.skip_list = []

    def __repr__(self):
        s = ("%(type)s instance --\n"
             "device = %(device)s  length = %(length)d  size = %(size)s\n"
             "remaining = %(rem)d  pool = %(pool)d" %
             {"type": self.__class__.__name__, "device": self.path,
              "length": self.length, "size": self.length_to_size(self.length),
              "pool": self.pool, "rem": self.remaining})

        return s

    # Force str and unicode types in case path is unicode
    def _to_string(self):
        s = "%d on %s" % (self.length, self.path)
        return s

    def __str__(self):
        return stringize(self._to_string())

    def __unicode__(self):
        return unicodeize(self._to_string())

    def add_request(self, req):
        """ Add a request to this chunk.

            :param req: the request to add
            :type req: :class:`Request`
        """
        log.debug("adding request %d to chunk %s", req.device.id, self)

        self.requests.append(req)
        self.pool -= req.base
        self.pool -= req.reserve_request

        if not req.done:
            self.base += req.base

    def reclaim(self, request, amount):
        """ Reclaim units from a request and return them to the pool.

            :param request: the request to reclaim units from
            :type request: :class:`Request`
            :param amount: number of units to reclaim from the request
            :type amount: int
            :raises: ValueError
            :returns: None
        """
        log.debug("reclaim: %s %d (%s)", request, amount, self.length_to_size(amount))
        if request.growth < amount:
            log.error("tried to reclaim %d from request with %d of growth",
                      amount, request.growth)
            raise ValueError(_("cannot reclaim more than request has grown"))

        request.growth -= amount
        self.pool += amount

        # put this request in the skip list so we don't try to grow it the
        # next time we call grow_requests to allocate the newly re-acquired pool
        if request not in self.skip_list:
            self.skip_list.append(request)

    @property
    def growth(self):
        """ Sum of growth for all requests in this chunk. """
        return sum(r.growth for r in self.requests)

    @property
    def has_growable(self):
        """ True if this chunk contains at least one growable request. """
        for req in self.requests:
            if req.growable:
                return True
        return False

    @property
    def remaining(self):
        """ Number of requests still being grown in this chunk. """
        return len([d for d in self.requests if not d.done])

    @property
    def done(self):
        """ True if we are finished growing all requests in this chunk. """
        return self.remaining == 0 or self.pool == 0

    def max_growth(self, req):
        return req.max_growth

    def length_to_size(self, length):
        return length

    def size_to_length(self, size):
        return size

    def trim_over_grown_request(self, req, base=None):
        """ Enforce max growth and return extra units to the pool.

            :param req: the request to trim
            :type req: :class:`Request`
            :keyword base: base unit count to adjust if req is done growing
            :type base: int
            :returns: the new base or None if no base was given
            :rtype: int or None
        """
        max_growth = self.max_growth(req)
        if max_growth and req.growth >= max_growth:
            if req.growth > max_growth:
                # we've grown beyond the maximum. put some back.
                extra = req.growth - max_growth
                log.debug("taking back %d (%s) from %d (%s)",
                          extra, self.length_to_size(extra),
                          req.device.id, req.device.name)
                self.pool += extra
                req.growth = max_growth

            # We're done growing this request, so it no longer
            # factors into the growable base used to determine
            # what fraction of the pool each request gets.
            if base is not None:
                base -= req.base
            req.done = True

        return base

    def sort_requests(self):
        pass

    def grow_requests(self, uniform=False):
        """ Calculate growth amounts for requests in this chunk.

            :keyword uniform: grow requests uniformly instead of proportionally
            :type uniform: bool

            The default mode of growth is as follows: given a total number of
            available units, requests receive an allotment proportional to their
            base sizes. That means a request with base size 1000 will grow four
            times as fast as a request with base size 250.

            Under uniform growth, all requests receive an equal portion of the
            free units.
        """
        log.debug("Chunk.grow_requests: %r", self)

        self.sort_requests()
        for req in self.requests:
            log.debug("req: %r", req)

        # we use this to hold the base for the next loop through the
        # chunk's requests since we want the base to be the same for
        # all requests in any given growth iteration
        new_base = self.base
        last_pool = 0  # used to track changes to the pool across iterations
        while not self.done and self.pool and last_pool != self.pool:
            last_pool = self.pool    # to keep from getting stuck
            self.base = new_base
            if uniform:
                growth = int(last_pool / self.remaining)

            log.debug("%d requests and %s (%s) left in chunk",
                      self.remaining, self.pool, self.length_to_size(self.pool))
            for p in self.requests:
                if p.done or p in self.skip_list:
                    continue

                if not uniform:
                    # Each request is allocated free units from the pool
                    # based on the relative _base_ sizes of the remaining
                    # growable requests.
                    share = Decimal(p.base) / Decimal(self.base)
                    growth = int(share * last_pool)  # truncate, don't round

                p.growth += growth
                self.pool -= growth
                log.debug("adding %s (%s) to %d (%s)",
                          growth, self.length_to_size(growth),
                          p.device.id, p.device.name)

                new_base = self.trim_over_grown_request(p, base=new_base)
                log.debug("new grow amount for request %d (%s) is %s "
                          "units, or %s",
                          p.device.id, p.device.name, p.growth,
                          self.length_to_size(p.growth))

        if self.pool:
            # allocate any leftovers in pool to the first partition
            # that can still grow
            for p in self.requests:
                if p.done or p in self.skip_list:
                    continue

                growth = self.pool
                p.growth += growth
                self.pool = 0
                log.debug("adding %s (%s) to %d (%s)",
                          growth, self.length_to_size(growth),
                          p.device.id, p.device.name)

                self.trim_over_grown_request(p)
                log.debug("new grow amount for request %d (%s) is %s "
                          "units, or %s",
                          p.device.id, p.device.name, p.growth,
                          self.length_to_size(p.growth))

                if self.pool == 0:
                    break

        # requests that were skipped over this time through are back on the
        # table next time
        self.skip_list = []


class DiskChunk(Chunk):

    """ A free region on disk from which partitions will be allocated """

    def __init__(self, geometry, requests=None):
        """
            :param geometry: the free region this chunk represents
            :type geometry: :class:`parted.Geometry`
            :keyword requests: list of requests to add initially
            :type requests: list of :class:`PartitionRequest`

            .. note::

                We will limit partition growth based on disklabel limitations
                for partition end sector, so a 10TB disk with an msdos disklabel
                will be treated like a 2TiB disk.

            .. note::

                If you plan to allocate aligned partitions you should pass in an
                aligned geometry instance.

        """
        self.geometry = geometry            # parted.Geometry
        self.sector_size = Size(self.geometry.device.sectorSize)
        self.path = self.geometry.device.path
        super(DiskChunk, self).__init__(self.geometry.length, requests=requests)

    def __repr__(self):
        s = super(DiskChunk, self).__str__()
        s += (" start = %(start)d  end = %(end)d\n"
              "sector_size = %(sector_size)s\n" %
              {"start": self.geometry.start, "end": self.geometry.end,
               "sector_size": self.sector_size})
        return s

    # Force str and unicode types in case path is unicode
    def _to_string(self):
        s = "%d (%d-%d) on %s" % (self.length, self.geometry.start,
                                  self.geometry.end, self.path)
        return s

    def __str__(self):
        return stringize(self._to_string())

    def __unicode__(self):
        return unicodeize(self._to_string())

    def add_request(self, req):
        """ Add a request to this chunk.

            :param req: the request to add
            :type req: :class:`PartitionRequest`
        """
        if not isinstance(req, PartitionRequest):
            raise ValueError(_("DiskChunk requests must be of type "
                               "PartitionRequest"))

        if not self.requests:
            # when adding the first request to the chunk, adjust the pool
            # size to reflect any disklabel-specific limits on end sector
            max_sector = req.device.parted_partition.disk.maxPartitionStartSector
            chunk_end = min(max_sector, self.geometry.end)
            if chunk_end <= self.geometry.start:
                # this should clearly never be possible, but if the chunk's
                # start sector is beyond the maximum allowed end sector, we
                # cannot continue
                log.error("chunk start sector is beyond disklabel maximum")
                raise PartitioningError(_("partitions allocated outside "
                                          "disklabel limits"))

            new_pool = chunk_end - self.geometry.start + 1
            if new_pool != self.pool:
                log.debug("adjusting pool to %d based on disklabel limits", new_pool)
                self.pool = new_pool

        super(DiskChunk, self).add_request(req)

    def max_growth(self, req):
        """ Return the maximum possible growth for a request.

            :param req: the request
            :type req: :class:`PartitionRequest`
        """
        req_end = req.device.parted_partition.geometry.end
        req_start = req.device.parted_partition.geometry.start

        # Establish the current total number of sectors of growth for requests
        # that lie before this one within this chunk. We add the total count
        # to this request's end sector to obtain the end sector for this
        # request, including growth of earlier requests but not including
        # growth of this request. Maximum growth values are obtained using
        # this end sector and various values for maximum end sector.
        growth = 0
        for request in self.requests:
            if request.device.parted_partition.geometry.start < req_start:
                growth += request.growth
        req_end += growth

        # obtain the set of possible maximum sectors-of-growth values for this
        # request and use the smallest
        limits = []

        # disklabel-specific maximum sector
        max_sector = req.device.parted_partition.disk.maxPartitionStartSector
        limits.append(max_sector - req_end)

        # 2TB limit on bootable partitions, regardless of disklabel
        if req.device.req_bootable:
            max_boot = size_to_sectors(Size("2 TiB"), self.sector_size)
            limits.append(max_boot - req_end)

        # request-specific maximum (see Request.__init__, above, for details)
        if req.max_growth:
            limits.append(req.max_growth)

        max_growth = min(limits)
        return max_growth

    def length_to_size(self, length):
        return sectors_to_size(length, self.sector_size)

    def size_to_length(self, size):
        return size_to_sectors(size, self.sector_size)

    def sort_requests(self):
        # sort the partitions by start sector
        self.requests.sort(key=lambda r: r.device.parted_partition.geometry.start)


class VGChunk(Chunk):

    """ A free region in an LVM VG from which LVs will be allocated """

    def __init__(self, vg, requests=None):
        """
            :param vg: the volume group whose free space this chunk represents
            :type vg: :class:`~.devices.LVMVolumeGroupDevice`
            :keyword requests: list of requests to add initially
            :type requests: list of :class:`LVRequest`
        """
        self.vg = vg
        self.path = vg.path
        usable_extents = vg.extents - int(vg.align(vg.reserved_space, roundup=True) / vg.pe_size)
        super(VGChunk, self).__init__(usable_extents, requests=requests)

    def add_request(self, req):
        """ Add a request to this chunk.

            :param req: the request to add
            :type req: :class:`LVRequest`
        """
        if not isinstance(req, LVRequest):
            raise ValueError(_("VGChunk requests must be of type "
                               "LVRequest"))

        super(VGChunk, self).add_request(req)

    def length_to_size(self, length):
        return self.vg.pe_size * length

    def size_to_length(self, size):
        return int(size / self.vg.pe_size)

    def sort_requests(self):
        # sort the partitions by start sector
        self.requests.sort(key=_lv_compare_key)


class ThinPoolChunk(VGChunk):

    """ A free region in an LVM thin pool from which LVs will be allocated """

    def __init__(self, pool, requests=None):
        """
            :param pool: the thin pool whose free space this chunk represents
            :type pool: :class:`~.devices.LVMLogicalVolumeDevice`
            :keyword requests: list of requests to add initially
            :type requests: list of :class:`LVRequest`
        """
        self.vg = pool.vg   # only used for align, &c
        self.path = pool.path
        usable_extents = (pool.size / pool.vg.pe_size)
        super(VGChunk, self).__init__(usable_extents, requests=requests)  # pylint: disable=bad-super-call


def get_disk_chunks(disk, partitions, free):
    """ Return a list of Chunk instances representing a disk.

        :param disk: the disk
        :type disk: :class:`~.devices.StorageDevice`
        :param partitions: list of partitions
        :type partitions: list of :class:`~.devices.PartitionDevice`
        :param free: list of free regions
        :type free: list of :class:`parted.Geometry`
        :returns: list of chunks representing the disk
        :rtype: list of :class:`DiskChunk`

        Partitions and free regions not on the specified disk are ignored.

        Chunks contain an aligned version of the free region's geometry.
    """
    # list of all new partitions on this disk
    disk_parts = [p for p in partitions if p.disk == disk and not p.exists]
    disk_free = [f for f in free if f.device.path == disk.path]

    chunks = []
    for f in disk_free[:]:
        # Align the geometry so we have a realistic view of the free space.
        # alignUp and alignDown can align in the reverse direction if the only
        # aligned sector within the geometry is in that direction, so we have to
        # also check that the resulting aligned geometry has a non-zero length.
        # (It is possible that both will align to the same sector in a small
        #  enough region.)
        try:
            size = sectors_to_size(f.length, disk.format.sector_size)
            alignment = disk.format.get_alignment(size=size)
            end_alignment = disk.format.get_end_alignment(alignment=alignment)
        except AlignmentError:
            disk_free.remove(f)
            continue

        al_start = alignment.alignUp(f, f.start)
        al_end = end_alignment.alignDown(f, f.end)
        if al_start >= al_end:
            disk_free.remove(f)
            continue
        geom = parted.Geometry(device=f.device,
                               start=al_start,
                               end=al_end)
        if geom.length < alignment.grainSize:
            disk_free.remove(f)
            continue

        chunks.append(DiskChunk(geom))

    for p in disk_parts:
        if p.is_extended:
            # handle extended partitions specially since they are
            # indeed very special
            continue

        for i, f in enumerate(disk_free):
            if f.contains(p.parted_partition.geometry):
                chunks[i].add_request(PartitionRequest(p))
                break

    return chunks


class TotalSizeSet(object):

    """ Set of device requests with a target combined size.

        This will be handled by growing the requests until the desired combined
        size has been achieved.
    """

    def __init__(self, devices, size):
        """
            :param devices: the set of devices
            :type devices: list of :class:`~.devices.PartitionDevice`
            :param size: the target combined size
            :type size: :class:`~.size.Size`
        """
        self.devices = [d.raw_device for d in devices]
        self.size = size

        self.requests = []

        self.allocated = sum((d.req_base_size for d in self.devices), Size(0))
        log.debug("set.allocated = %d", self.allocated)

    def allocate(self, amount):
        log.debug("allocating %d to TotalSizeSet with %d/%d (%d needed)",
                  amount, self.allocated, self.size, self.needed)
        self.allocated += amount

    @property
    def needed(self):
        return self.size - self.allocated

    def deallocate(self, amount):
        log.debug("deallocating %d from TotalSizeSet with %d/%d (%d needed)",
                  amount, self.allocated, self.size, self.needed)
        self.allocated -= amount


class SameSizeSet(object):

    """ Set of device requests with a common target size. """

    def __init__(self, devices, size, grow=False, max_size=None):
        """
            :param devices: the set of devices
            :type devices: list of :class:`~.devices.PartitionDevice`
            :param size: target size for each device/request
            :type size: :class:`~.size.Size`
            :keyword grow: whether the devices can be grown
            :type grow: bool
            :keyword max_size: the maximum size for growable devices
            :type max_size: :class:`~.size.Size`
        """
        self.devices = [d.raw_device for d in devices]
        self.size = size / len(devices)
        self.grow = grow
        self.max_size = max_size

        self.requests = []


def manage_size_sets(size_sets, chunks):
    growth_by_request = {}
    requests_by_device = {}
    chunks_by_request = {}
    for chunk in chunks:
        for request in chunk.requests:
            requests_by_device[request.device] = request
            chunks_by_request[request] = chunk
            growth_by_request[request] = 0

    for i in range(2):
        reclaimed = dict([(chunk, 0) for chunk in chunks])
        for ss in size_sets:
            if isinstance(ss, TotalSizeSet):
                # TotalSizeSet members are trimmed to achieve the requested
                # total size
                log.debug("set: %s %d/%d", [d.name for d in ss.devices],
                          ss.allocated, ss.size)

                for device in ss.devices:
                    request = requests_by_device[device]
                    chunk = chunks_by_request[request]
                    new_growth = request.growth - growth_by_request[request]
                    ss.allocate(chunk.length_to_size(new_growth))

                # decide how much to take back from each request
                # We may assume that all requests have the same base size.
                # We're shooting for a roughly equal distribution by trimming
                # growth from the requests that have grown the most first.
                requests = sorted([requests_by_device[d] for d in ss.devices],
                                  key=lambda r: r.growth, reverse=True)
                needed = ss.needed
                for request in requests:
                    chunk = chunks_by_request[request]
                    log.debug("%s", request)
                    log.debug("needed: %d", ss.needed)

                    if ss.needed < 0:
                        # it would be good to take back some from each device
                        # instead of taking all from the last one(s)
                        extra = -chunk.size_to_length(needed) // len(ss.devices)
                        if extra > request.growth and i == 0:
                            log.debug("not reclaiming from this request")
                            continue
                        else:
                            extra = min(extra, request.growth)

                        reclaimed[chunk] += extra
                        chunk.reclaim(request, extra)
                        ss.deallocate(chunk.length_to_size(extra))

                    if ss.needed <= 0:
                        request.done = True

            elif isinstance(ss, SameSizeSet):
                # SameSizeSet members all have the same size as the smallest
                # member
                requests = [requests_by_device[d] for d in ss.devices]
                _min_growth = min([r.growth for r in requests])
                log.debug("set: %s %d", [d.name for d in ss.devices], ss.size)
                log.debug("min growth is %d", _min_growth)
                for request in requests:
                    chunk = chunks_by_request[request]
                    _max_growth = chunk.size_to_length(ss.size) - request.base
                    log.debug("max growth for %s is %d", request, _max_growth)
                    min_growth = max(min(_min_growth, _max_growth), 0)
                    if request.growth > min_growth:
                        extra = request.growth - min_growth
                        reclaimed[chunk] += extra
                        chunk.reclaim(request, extra)
                        request.done = True
                    elif request.growth == min_growth:
                        request.done = True

        # store previous growth amounts so we know how much was allocated in
        # the latest grow_requests call
        for request in growth_by_request.keys():
            growth_by_request[request] = request.growth

        for chunk in chunks:
            if reclaimed[chunk] and not chunk.done:
                chunk.grow_requests()


def grow_partitions(disks, partitions, free, size_sets=None):
    """ Grow all growable partition requests.

        Partitions have already been allocated from chunks of free space on
        the disks. This function does not modify the ordering of partitions
        or the free chunks from which they are allocated.

        Free space within a given chunk is allocated to each growable
        partition allocated from that chunk in an amount corresponding to
        the ratio of that partition's base size to the sum of the base sizes
        of all growable partitions allocated from the chunk.

        :param disks: all usable disks
        :type disks: list of :class:`~.devices.StorageDevice`
        :param partitions: all partitions
        :type partitions: list of :class:`~.devices.PartitionDevice`
        :param free: all free regions on disks
        :type free: list of :class:`parted.Geometry`
        :keyword size_sets: list of size-related partition sets
        :type size_sets: list of :class:`TotalSizeSet` or :class:`SameSizeSet`
        :returns: :const:`None`
    """
    log.debug("grow_partitions: disks=%s, partitions=%s",
              [d.name for d in disks],
              ["%s(id %d)" % (p.name, p.id) for p in partitions])
    all_growable = [p for p in partitions if p.req_grow]
    if not all_growable:
        log.debug("no growable partitions")
        return

    if size_sets is None:
        size_sets = []

    log.debug("growable partitions are %s", [p.name for p in all_growable])

    #
    # collect info about each disk and the requests it contains
    #
    chunks = []
    for disk in disks:
        # list of free space regions on this disk prior to partition allocation
        disk_free = [f for f in free if f.device.path == disk.path]
        if not disk_free:
            log.debug("no free space on %s", disk.name)
            continue

        disk_chunks = get_disk_chunks(disk, partitions, disk_free)
        log.debug("disk %s has %d chunks", disk.name, len(disk_chunks))
        chunks.extend(disk_chunks)

    #
    # grow the partitions in each chunk as a group
    #
    for chunk in chunks:
        if not chunk.has_growable:
            # no growable partitions in this chunk
            continue

        chunk.grow_requests()

    # adjust set members' growth amounts as needed
    manage_size_sets(size_sets, chunks)

    for disk in disks:
        log.debug("growing partitions on %s", disk.name)
        for chunk in chunks:
            if chunk.path != disk.path:
                continue

            if not chunk.has_growable:
                # no growable partitions in this chunk
                continue

            # recalculate partition geometries
            disklabel = disk.format
            start = chunk.geometry.start
            default_alignment = disklabel.get_alignment()

            # find any extended partition on this disk
            extended_geometry = getattr(disklabel.extended_partition,
                                        "geometry",
                                        None)  # parted.Geometry

            # align start sector as needed
            if not default_alignment.isAligned(chunk.geometry, start):
                start = default_alignment.alignUp(chunk.geometry, start)
            new_partitions = []
            for p in chunk.requests:
                ptype = p.device.parted_partition.type
                log.debug("partition %s (%d): %s", p.device.name,
                          p.device.id, ptype)
                if ptype == parted.PARTITION_EXTENDED:
                    continue

                new_length = p.base + p.growth
                alignment = disklabel.get_alignment(size=chunk.length_to_size(new_length))
                end_alignment = disklabel.get_end_alignment(alignment=alignment)
                # XXX since we need one metadata sector before each
                #     logical partition we burn one logical block to
                #     safely align the start of each logical partition
                if ptype == parted.PARTITION_LOGICAL:
                    start += alignment.grainSize

                end = start + new_length - 1
                # align end sector as needed
                if not end_alignment.isAligned(chunk.geometry, end):
                    end = end_alignment.alignDown(chunk.geometry, end)
                new_geometry = parted.Geometry(device=disklabel.parted_device,
                                               start=start,
                                               end=end)
                log.debug("new geometry for %s: %s", p.device.name,
                          new_geometry)
                start = end + 1
                new_partition = parted.Partition(disk=disklabel.parted_disk,
                                                 type=ptype,
                                                 geometry=new_geometry)
                new_partitions.append((new_partition, p.device))

            # remove all new partitions from this chunk
            remove_new_partitions([disk], [r.device for r in chunk.requests],
                                  partitions)
            log.debug("back from remove_new_partitions")

            # adjust the extended partition as needed
            # we will ony resize an extended partition that we created
            log.debug("extended: %s", extended_geometry)
            if extended_geometry and \
               chunk.geometry.contains(extended_geometry):
                log.debug("setting up new geometry for extended on %s", disk.name)
                ext_start = 0
                for (partition, device) in new_partitions:
                    if partition.type != parted.PARTITION_LOGICAL:
                        continue

                    if not ext_start or partition.geometry.start < ext_start:
                        # account for the logical block difference in start
                        # sector for the extended -v- first logical
                        # (partition.geometry.start is already aligned)
                        ext_start = partition.geometry.start - default_alignment.grainSize

                new_geometry = parted.Geometry(device=disklabel.parted_device,
                                               start=ext_start,
                                               end=chunk.geometry.end)
                log.debug("new geometry for extended: %s", new_geometry)
                new_extended = parted.Partition(disk=disklabel.parted_disk,
                                                type=parted.PARTITION_EXTENDED,
                                                geometry=new_geometry)
                ptypes = [p.type for (p, d) in new_partitions]
                for pt_idx, ptype in enumerate(ptypes):
                    if ptype == parted.PARTITION_LOGICAL:
                        new_partitions.insert(pt_idx, (new_extended, None))
                        break

            # add the partitions with their new geometries to the disk
            for (partition, device) in new_partitions:
                if device:
                    name = device.name
                else:
                    # If there was no extended partition on this disk when
                    # do_partitioning was called we won't have a
                    # PartitionDevice instance for it.
                    name = partition.getDeviceNodeName()

                log.debug("setting %s new geometry: %s", name,
                          partition.geometry)
                constraint = parted.Constraint(exactGeom=partition.geometry)
                disklabel.parted_disk.addPartition(partition=partition,
                                                   constraint=constraint)
                path = partition.path
                if device:
                    # set the device's name
                    device.parted_partition = partition
                    # without this, the path attr will be a basename. eek.
                    device.disk = disk

                    # make sure we store the disk's version of the partition
                    newpart = disklabel.parted_disk.getPartitionByPath(path)
                    device.parted_partition = newpart


def lv_compare(lv1, lv2):
    """ More specifically defined lvs come first.

        < 1 => x < y
          0 => x == y
        > 1 => x > y
    """
    if not isinstance(lv1, Device):
        lv1 = lv1.device
    if not isinstance(lv2, Device):
        lv2 = lv2.device

    ret = 0

    # larger requests go to the front of the list
    ret -= compare(lv1.size, lv2.size) * 100

    # fixed size requests to the front
    ret += compare(lv1.req_grow, lv2.req_grow) * 50

    # potentially larger growable requests go to the front
    if lv1.req_grow and lv2.req_grow:
        if not lv1.req_max_size and lv2.req_max_size:
            ret -= 25
        elif lv1.req_max_size and not lv2.req_max_size:
            ret += 25
        else:
            ret -= compare(lv1.req_max_size, lv2.req_max_size) * 25

    if ret > 0:
        ret = 1
    elif ret < 0:
        ret = -1

    return ret


_lv_compare_key = functools.cmp_to_key(lv_compare)


def _apply_chunk_growth(chunk):
    """ grow the lvs by the amounts the VGChunk calculated """
    for req in chunk.requests:
        if not req.device.req_grow:
            continue

        size = chunk.length_to_size(req.base + req.growth)

        # Base is pe, which means potentially rounded up by as much as
        # pesize-1. As a result, you can't just add the growth to the
        # initial size.
        req.device.size = size


def grow_lvm(storage):
    """ Grow LVs according to the sizes of the PVs.

        Strategy for growth involving thin pools:
            - Applies to device factory class as well.
            - Overcommit is not allowed.
            - Pool lv's base size includes sizes of thin lvs within it.
            - Pool is grown along with other non-thin lvs.
            - Thin lvs within each pool are grown separately using the
              ThinPoolChunk class.
    """
    for vg in storage.vgs:
        total_free = vg.free_space
        if total_free < 0:
            # by now we have allocated the PVs so if there isn't enough
            # space in the VG we have a real problem
            raise PartitioningError(_("not enough space for LVM requests"))
        elif not total_free:
            log.debug("vg %s has no free space", vg.name)
            continue

        log.debug("vg %s: %s free ; lvs: %s", vg.name, total_free,
                  [l.lvname for l in vg.lvs])

        # don't include thin lvs in the vg's growth calculation
        fatlvs = [lv for lv in vg.lvs if lv not in vg.thinlvs]
        requests = []
        for lv in fatlvs:
            if lv in vg.thinpools:
                # make sure the pool's base size is at least the sum of its lvs'
                lv.req_size = max(lv.min_size, lv.req_size, lv.used_space)
                lv.size = lv.req_size

        # establish sizes for the percentage-based requests (which are fixed)
        percentage_based_lvs = [lv for lv in vg.lvs if lv.req_percent]
        if sum(lv.req_percent for lv in percentage_based_lvs) > 100:
            raise ValueError("sum of percentages within a vg cannot exceed 100")

        percent_base = sum(vg.align(lv.req_size, roundup=False) / vg.pe_size
                           for lv in percentage_based_lvs)
        percentage_basis = vg.free_extents + percent_base
        for lv in percentage_based_lvs:
            new_extents = int(lv.req_percent * Decimal('0.01') * percentage_basis)
            # set req_size also so the request can also be growable if desired
            lv.size = lv.req_size = vg.pe_size * new_extents

        # grow regular lvs
        chunk = VGChunk(vg, requests=[LVRequest(l) for l in fatlvs])
        chunk.grow_requests()
        _apply_chunk_growth(chunk)

        # now that we have grown all thin pools (if any), let's calculate and
        # set their metadata size if not told otherwise
        for pool in vg.thinpools:
            orig_pmspare_size = vg.pmspare_size
            if not pool.exists and pool.metadata_size == Size(0):
                pool.autoset_md_size()
            if vg.pmspare_size != orig_pmspare_size:
                # pmspare size change caused by the above step, let's trade part
                # of pool's space for it
                pool.size -= vg.pmspare_size - orig_pmspare_size

        # now, grow thin lv requests within their respective pools
        for pool in vg.thinpools:
            requests = [LVRequest(l) for l in pool.lvs]
            thin_chunk = ThinPoolChunk(pool, requests)
            thin_chunk.grow_requests()
            _apply_chunk_growth(thin_chunk)



#
# Copyright (C) 2013-2015  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): Vratislav Podzimek <vpodzime@redhat.com>
#
# This module was taken from the sources of the Anaconda installer.
#

"""Module providing thread-safe and mainloop-safe DBus operations."""

import gi
gi.require_version("GLib", "2.0")
gi.require_version("Gio", "2.0")

from gi.repository import GLib, Gio

import os

DEFAULT_DBUS_TIMEOUT = -1
DBUS_PROPS_IFACE = "org.freedesktop.DBus.Properties"
DBUS_INTRO_IFACE = "org.freedesktop.DBus.Introspectable"


class SafeDBusError(Exception):
    """Class for exceptions defined in this module."""


class DBusCallError(SafeDBusError):
    """Class for the errors related to calling methods over DBus."""


class DBusPropertyError(DBusCallError):
    """Class for the errors related to getting property values over DBus."""


def get_new_system_connection():
    """Return a new connection to the system bus."""

    return Gio.DBusConnection.new_for_address_sync(
        Gio.dbus_address_get_for_bus_sync(Gio.BusType.SYSTEM, None),
        Gio.DBusConnectionFlags.AUTHENTICATION_CLIENT |
        Gio.DBusConnectionFlags.MESSAGE_BUS_CONNECTION,
        None, None)


def get_new_session_connection():
    """
    Get a connection handle for the per-user-login-session message bus.

    !!! RUN THIS EARLY !!! like, before any other threads start. Connections to
    the session bus must be made with the effective UID of the login user,
    which in live installs is not the UID of anaconda. This means we need to
    call seteuid in this method, and doing that after threads have started will
    probably do something weird.

    Live installs use consolehelper to run as root, which sets the original
    UID in $USERHELPER_UID.

    :return: the session connection handle
    :rtype: Gio.DBusConnection
    :raise DBusCallError: if some DBus related error appears
    :raise OSError: if unable to set the effective UID
    """

    old_euid = None
    if "USERHELPER_UID" in os.environ:
        old_euid = os.geteuid()
        os.seteuid(int(os.environ["USERHELPER_UID"]))

    try:
        connection = Gio.DBusConnection.new_for_address_sync(
            Gio.dbus_address_get_for_bus_sync(Gio.BusType.SESSION, None),
            Gio.DBusConnectionFlags.AUTHENTICATION_CLIENT |
            Gio.DBusConnectionFlags.MESSAGE_BUS_CONNECTION,
            None, None)
    except GLib.GError as gerr:
        raise DBusCallError("Unable to connect to session bus: %s" % gerr)
    finally:
        if old_euid is not None:
            os.seteuid(old_euid)

    if connection.is_closed():
        raise DBusCallError("Connection is closed")

    return connection


def call_sync(service, obj_path, iface, method, args,
              connection=None, fds=None):
    """
    Safely call a given method on a given object of a given service over DBus
    passing given arguments. If a connection is given, it is used, otherwise a
    new connection is established. Safely means that it is a synchronous,
    thread-safe call not using any main loop.

    :param service: DBus service to use
    :type service: str
    :param obj_path: object path of the object to call method on
    :type obj_path: str
    :param iface: interface to use
    :type iface: str
    :param method: name of the method to call
    :type method: str
    :param args: arguments to pass to the method
    :type args: GVariant
    :param connection: connection to use (if None, a new connection is
                       established)
    :type connection: Gio.DBusConnection
    :param fds: list of file descriptors for the call
    :type: Gio.UnixFDList
    :return: unpacked value returned by the method
    :rtype: tuple with elements that depend on the method
    :raise DBusCallError: if some DBus related error appears

    """

    if not connection:
        try:
            connection = get_new_system_connection()
        except GLib.GError as gerr:
            raise DBusCallError("Unable to connect to system bus: %s" % gerr)

    if connection.is_closed():
        raise DBusCallError("Connection is closed")

    try:
        ret = connection.call_with_unix_fd_list_sync(service, obj_path, iface, method, args,
                                                     None, Gio.DBusCallFlags.NONE,
                                                     DEFAULT_DBUS_TIMEOUT, fds, None)
    except GLib.GError as gerr:
        msg = "Failed to call %s method on %s with %s arguments: %s" % \
              (method, obj_path, args, gerr.message)  # pylint: disable=no-member
        raise DBusCallError(msg)

    if ret is None:
        msg = "No return from %s method on %s with %s arguments" % (method, obj_path, args)
        raise DBusCallError(msg)

    return ret[0].unpack()


def get_property_sync(service, obj_path, iface, prop_name,
                      connection=None):
    """
    Get value of a given property of a given object provided by a given service.

    :param service: DBus service to use
    :type service: str
    :param obj_path: object path
    :type obj_path: str
    :param iface: interface to use
    :type iface: str
    :param prop_name: name of the property
    :type prop_name: str
    :param connection: connection to use (if None, a new connection is
                       established)
    :type connection: Gio.DBusConnection
    :return: unpacked value of the property
    :rtype: tuple with elements that depend on the type of the property
    :raise DBusCallError: when the internal dbus_call_safe_sync invocation
                          raises an exception
    :raise DBusPropertyError: when the given object doesn't have the given
                              property

    """

    args = GLib.Variant('(ss)', (iface, prop_name))
    ret = call_sync(service, obj_path, DBUS_PROPS_IFACE, "Get", args,
                    connection)
    if ret is None:
        msg = "No value for the %s object's property %s" % (obj_path, prop_name)
        raise DBusPropertyError(msg)

    return ret


def get_properties_sync(service, obj_path, iface, connection=None):
    """
    Get all properties of a given object provided by a given service.

    :param service: DBus service to use
    :type service: str
    :param obj_path: object path
    :type obj_path: str
    :param iface: interface to use
    :type iface: str
    :param connection: connection to use (if None, a new connection is
                       established)
    :type connection: Gio.DBusConnection
    :return: unpacked value of the property
    :rtype: tuple with elements that depend on the type of the property
    :raise DBusCallError: when the internal dbus_call_safe_sync invocation
                          raises an exception

    """

    args = GLib.Variant('(s)', (iface,))
    ret = call_sync(service, obj_path, DBUS_PROPS_IFACE, "GetAll", args,
                    connection)

    return ret


def check_object_available(service, obj_path, iface=None):
    intro_data = call_sync(service, obj_path, DBUS_INTRO_IFACE, "Introspect", None)
    node_info = Gio.DBusNodeInfo.new_for_xml(intro_data[0])
    if not iface:
        # just check if any interface is available (there are none for
        # non-existing objects)
        return bool(node_info.interfaces)
    else:
        return any(intface.name == iface for intface in node_info.interfaces)



# devices/loop.py
#
# Copyright (C) 2009-2014  Red Hat, Inc.
#
# This copyrighted material is made available to anyone wishing to use,
# modify, copy, or redistribute it subject to the terms and conditions of
# the GNU General Public License v.2, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY expressed or implied, including the implied warranties of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General
# Public License for more details.  You should have received a copy of the
# GNU General Public License along with this program; if not, write to the
# Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.  Any Red Hat trademarks that are incorporated in the
# source code or documentation are not subject to the GNU General Public
# License and may only be used or replicated with the express permission of
# Red Hat, Inc.
#
# Red Hat Author(s): David Lehman <dlehman@redhat.com>
#

import gi
gi.require_version("BlockDev", "3.0")

from gi.repository import BlockDev as blockdev

import os

from .. import errors
from ..storage_log import log_method_call
from ..tasks import availability

import logging
log = logging.getLogger("blivet")

from .storage import StorageDevice


class LoopDevice(StorageDevice):

    """ A loop device. """
    _type = "loop"
    _external_dependencies = [availability.BLOCKDEV_LOOP_PLUGIN]

    def __init__(self, name=None, fmt=None, size=None, sysfs_path=None,
                 exists=False, parents=None):
        """
            :param name: the device name (generally a device node's basename)
            :type name: str
            :keyword exists: does this device exist?
            :type exists: bool
            :keyword size: the device's size
            :type size: :class:`~.size.Size`
            :keyword parents: a list of parent devices
            :type parents: list of :class:`StorageDevice`
            :keyword fmt: this device's formatting
            :type fmt: :class:`~.formats.DeviceFormat` or a subclass of it

            Loop devices always exist.
        """
        if not parents:
            raise ValueError("LoopDevice requires a backing device")

        if not name:
            # set up a temporary name until we've activated the loop device
            name = "tmploop%d" % self.id

        StorageDevice.__init__(self, name, fmt=fmt, size=size,
                               exists=True, parents=parents)

    def _set_name(self, value):
        self._name = value  # actual name is set by losetup

    def update_name(self):
        """ Update this device's name. """
        if not self.parents[0].status:
            # if the backing device is inactive, so are we
            return self.name

        if self.name.startswith("loop"):
            # if our name is loopN we must already be active
            return self.name

        try:
            name = blockdev.loop.get_loop_name(self.parents[0].path)
        except blockdev.LoopError as e:
            raise errors.LoopError(e)

        if name.startswith("loop"):
            self.name = name

        return self.name

    @property
    def status(self):
        return (self.parents[0].status and
                self.name.startswith("loop") and
                blockdev.loop.get_loop_name(self.parents[0].path) == self.name)

    @property
    def size(self):
        return self.parents[0].size

    def _pre_setup(self, orig=False):
        if not os.path.exists(self.parents[0].path):
            raise errors.DeviceError("specified file (%s) does not exist" % self.parents[0].path)
        return StorageDevice._pre_setup(self, orig=orig)

    def _setup(self, orig=False):
        """ Open, or set up, a device. """
        log_method_call(self, self.name, orig=orig, status=self.status,
                        controllable=self.controllable)
        try:
            blockdev.loop.setup(self.parents[0].path)
        except blockdev.LoopError as e:
            raise errors.LoopError(e)

    def _post_setup(self):
        StorageDevice._post_setup(self)
        self.update_name()
        self.update_sysfs_path()

    def _teardown(self, recursive=False):
        """ Close, or tear down, a device. """
        log_method_call(self, self.name, status=self.status,
                        controllable=self.controllable)
        try:
            blockdev.loop.teardown(self.path)
        except blockdev.LoopError as e:
            raise errors.LoopError(e)

    def _post_teardown(self, recursive=False):
        StorageDevice._post_teardown(self, recursive=recursive)
        self.name = "tmploop%d" % self.id
        self.sysfs_path = ''




import unittest
try:
    from unittest.mock import Mock, patch
except ImportError:
    from mock import Mock, patch

import parted

import blivet as blivet
from blivet.formats import get_format

# device classes for brevity's sake -- later on, that is
from blivet.devices import StorageDevice
from blivet.devices import PartitionDevice


class BlivetTestCase(unittest.TestCase):

    """ BlivetUnitTestCase

        This is a base class for unit test cases. It sets up imports of
        the blivet package. There are lots of little patches to prevent various
        pieces of code from trying to access filesystems and/or devices
        on the host system, along with a couple of convenience methods.

    """

    @patch("blivet.formats.fs.Ext4FS.supported", return_value=True)
    @patch("blivet.formats.fs.Ext4FS.formattable", return_value=True)
    def setUp(self, *args):  # pylint: disable=unused-argument,arguments-differ
        self.storage = blivet.Blivet()

        # device status
        self.storage_status = blivet.devices.StorageDevice.status
        self.dm_status = blivet.devices.DMDevice.status
        self.luks_status = blivet.devices.LUKSDevice.status
        self.vg_status = blivet.devices.LVMVolumeGroupDevice.status
        self.md_status = blivet.devices.MDRaidArrayDevice.status
        self.file_status = blivet.devices.FileDevice.status
        blivet.devices.StorageDevice.status = False
        blivet.devices.DMDevice.status = False
        blivet.devices.LUKSDevice.status = False
        blivet.devices.LVMVolumeGroupDevice.status = False
        blivet.devices.MDRaidArrayDevice.status = False
        blivet.devices.FileDevice.status = False

        # prevent PartitionDevice from trying to dig around in the partition's
        # geometry
        self.partition_set_target = PartitionDevice._set_target_size
        self.partition_align_target = PartitionDevice.align_target_size
        self.partition_max = PartitionDevice.max_size
        self.partition_min = PartitionDevice.min_size
        blivet.devices.PartitionDevice._set_target_size = StorageDevice._set_target_size
        blivet.devices.PartitionDevice.align_target_size = StorageDevice.align_target_size
        blivet.devices.PartitionDevice.max_size = StorageDevice.max_size
        blivet.devices.PartitionDevice.min_size = StorageDevice.min_size

        self.addCleanup(self._clean_up)

        def partition_probe(device):
            if isinstance(device._parted_partition, Mock):
                # don't clobber a Mock we already set up here
                part_mock = device._parted_partition
            else:
                part_mock = Mock()

            attrs = {"getLength.return_value": int(device._size),
                     "getDeviceNodeName.return_value": device.name,
                     "type": parted.PARTITION_NORMAL}
            part_mock.configure_mock(**attrs)
            device._parted_partition = part_mock
            device._current_size = device._size
            device._part_type = parted.PARTITION_NORMAL
            device._bootable = False

        self.partition_probe = PartitionDevice.probe
        PartitionDevice.probe = partition_probe

        self.get_active_mounts = blivet.formats.fs.mounts_cache._get_active_mounts
        blivet.formats.fs.mounts_cache._get_active_mounts = Mock()

    def _clean_up(self):
        blivet.devices.StorageDevice.status = self.storage_status
        blivet.devices.DMDevice.status = self.dm_status
        blivet.devices.LUKSDevice.status = self.luks_status
        blivet.devices.LVMVolumeGroupDevice.status = self.vg_status
        blivet.devices.MDRaidArrayDevice.status = self.md_status
        blivet.devices.FileDevice.status = self.file_status

        blivet.devices.PartitionDevice._set_target_size = self.partition_set_target
        blivet.devices.PartitionDevice.align_target_size = self.partition_align_target
        blivet.devices.PartitionDevice.max_size = self.partition_max
        blivet.devices.PartitionDevice.min_size = self.partition_min

        blivet.devices.PartitionDevice.probe = self.partition_probe

        blivet.formats.fs.mounts_cache._get_active_mounts = self.get_active_mounts

    def new_device(self, *args, **kwargs):
        """ Return a new Device instance suitable for testing. """
        device_class = kwargs.pop("device_class")

        # we intentionally don't pass the "exists" kwarg to the constructor
        # becauses this causes issues with some devices (especially partitions)
        # but we still need it for some LVs like VDO because we can't create
        # those so we need to fake their existence even for the constructor
        if device_class is blivet.devices.LVMLogicalVolumeDevice:
            exists = kwargs.get("exists", False)
        else:
            exists = kwargs.pop("exists", False)

        part_type = kwargs.pop("part_type", parted.PARTITION_NORMAL)
        device = device_class(*args, **kwargs)

        if exists:
            device._current_size = kwargs.get("size")

        if isinstance(device, blivet.devices.PartitionDevice):
            # if exists:
            #    device.parents = device.req_disks
            device.parents = device.req_disks

            parted_partition = Mock()

            if device.disk:
                part_num = device.name[len(device.disk.name):].split("p")[-1]
                parted_partition.number = int(part_num)

            parted_partition.type = part_type
            parted_partition.path = device.path
            parted_partition.get_device_node_name = Mock(return_value=device.name)
            if len(device.parents) == 1:
                disk_name = device.parents[0].name
                number = device.name.replace(disk_name, "")
                try:
                    parted_partition.number = int(number)
                except ValueError:
                    pass

            device._parted_partition = parted_partition
        elif isinstance(device, blivet.devices.LVMVolumeGroupDevice) and exists:
            device._complete = True

        device.exists = exists
        device.format.exists = exists

        if isinstance(device, blivet.devices.PartitionDevice):
            # PartitionDevice.probe sets up data needed for resize operations
            device.probe()

        return device

    def new_format(self, *args, **kwargs):
        """ Return a new DeviceFormat instance suitable for testing.

            Keyword Arguments:

                device_instance - StorageDevice instance this format will be
                                  created on. This is needed for setup of
                                  resizable formats.

            All other arguments are passed directly to
            blivet.formats.get_format.
        """
        exists = kwargs.pop("exists", False)
        device_instance = kwargs.pop("device_instance", None)
        fmt = get_format(*args, **kwargs)
        if isinstance(fmt, blivet.formats.disklabel.DiskLabel):
            fmt._parted_device = Mock()
            fmt._parted_disk = Mock()
            attrs = {"partitions": []}
            fmt._parted_disk.configure_mock(**attrs)

        fmt.exists = exists
        if exists:
            fmt._resizable = fmt.__class__._resizable

        if fmt.resizable and device_instance:
            fmt._size = device_instance.current_size

        return fmt

    def destroy_all_devices(self, disks=None):
        """ Remove all devices from the devicetree.

            Keyword Arguments:

                disks - a list of names of disks to remove partitions from

            Note: this is largely ripped off from partitioning.clear_partitions.

        """
        partitions = self.storage.partitions

        # Sort partitions by descending partition number to minimize confusing
        # things like multiple "destroy sda5" actions due to parted renumbering
        # partitions. This can still happen through the UI but it makes sense to
        # avoid it where possible.
        partitions.sort(key=lambda p: p.parted_partition.number, reverse=True)
        for part in partitions:
            if disks and part.disk.name not in disks:
                continue

            devices = self.storage.device_deps(part)
            while devices:
                leaves = [d for d in devices if d.isleaf]
                for leaf in leaves:
                    self.storage.destroy_device(leaf)
                    devices.remove(leaf)

            self.storage.destroy_device(part)

    def schedule_create_device(self, device):
        """ Schedule an action to create the specified device.

            Verify that the device is not already in the tree and that the
            act of scheduling/registering the action also adds the device to
            the tree.

            Return the DeviceAction instance.
        """
        if hasattr(device, "req_disks") and \
           len(device.req_disks) == 1 and \
           not device.parents:
            device.parents = device.req_disks

        devicetree = self.storage.devicetree

        self.assertEqual(devicetree.get_device_by_name(device.name), None)
        action = blivet.deviceaction.ActionCreateDevice(device)
        devicetree.actions.add(action)
        self.assertEqual(devicetree.get_device_by_name(device.name), device)
        return action

    def schedule_destroy_device(self, device):
        """ Schedule an action to destroy the specified device.

            Verify that the device exists initially and that the act of
            scheduling/registering the action also removes the device from
            the tree.

            Return the DeviceAction instance.
        """
        devicetree = self.storage.devicetree

        self.assertEqual(devicetree.get_device_by_name(device.name), device)
        action = blivet.deviceaction.ActionDestroyDevice(device)
        devicetree.actions.add(action)
        self.assertEqual(devicetree.get_device_by_name(device.name), None)
        return action

    def schedule_create_format(self, device, fmt):
        """ Schedule an action to write a new format to a device.

            Verify that the device is already in the tree, that it is not
            already set up to contain the specified format, and that the act
            of registering/scheduling the action causes the new format to be
            reflected in the tree.

            Return the DeviceAction instance.
        """
        devicetree = self.storage.devicetree

        self.assertNotEqual(device.format, fmt)
        self.assertEqual(devicetree.get_device_by_name(device.name), device)
        action = blivet.deviceaction.ActionCreateFormat(device, fmt)
        devicetree.actions.add(action)
        _device = devicetree.get_device_by_name(device.name)
        self.assertEqual(_device.format, fmt)
        return action

    def schedule_destroy_format(self, device):
        """ Schedule an action to remove a format from a device.

            Verify that the device is already in the tree and that the act
            of registering/scheduling the action causes the new format to be
            reflected in the tree.

            Return the DeviceAction instance.
        """
        devicetree = self.storage.devicetree

        self.assertEqual(devicetree.get_device_by_name(device.name), device)
        action = blivet.deviceaction.ActionDestroyFormat(device)
        devicetree.actions.add(action)
        _device = devicetree.get_device_by_name(device.name)
        self.assertEqual(_device.format.type, None)
        return action
